#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.slidejuncs`"""
import tempfile
import shutil
import os
import itertools
import numpy
import copy
import shlex

from nose.tools import assert_set_equal
from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.ref_files import REF_FILES
from plastid.util.io.filters import CommentReader
from plastid.readers.bed import BED_Reader
from plastid.genomics.roitools import SegmentChain
from plastid.bin.slidejuncs import main
from plastid.util.services.decorators import catch_stderr, catch_stdout

#===============================================================================
# INDEX: global constants
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.slidejuncs",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="slidejuncs"),
}

OUTBASE = os.path.join(test_info["temp_file_path"], "slidejuncs_%s_%s_%s")

# reference files used
# yapf: disable
FASTA_FILE   = REF_FILES["slidejuncs_seqs"]
COMBINED_BED = REF_FILES["slidejuncs_input"]
REF_FILE     = REF_FILES["slidejuncs_ref"]
MASK_FILE    = REF_FILES["slidejuncs_crossmap"]
# yapf: enable

#===============================================================================
# INDEX: helper functions
#===============================================================================


def get_junction_names_from_file(filename):
    """Return a set of names of splice junctions in a file

    Parameters
    ----------
    filename : str
        Fully-qualified path to file

    Returns
    -------
    set
        Set of splice junction names found in ``filename``
    """
    with open(filename) as fh:
        return set(
            [
                X.spanning_segment.chrom
                for X in BED_Reader(CommentReader(fh), return_type=SegmentChain)
            ]
        )


def get_junction_names_from_category(cat):
    """Helper function to deduce names of splice junctions/genes in each
    junction category from input files

    Parameters
    ----------
    cat : str
        Category of splice junction. Must be present in ``junction_categories``

    Returns
    -------
    set
        Set of junction names in input splice junction category ``cat``
    """
    return get_junction_names_from_file(REF_FILES["slidejuncs_%s" % cat])


#===============================================================================
# INDEX: programmatic definition of tests and expected results
#===============================================================================

# query junctions
junction_categories = [
    "known_juncs_non_crossmap",
    "known_juncs_crossmap",
    "to_slide_known_non_crossmap",
    "to_slide_known_crossmap",
    "noncan_no_ref",
    "expected_untouched",
]
"""Categories of splice junctions taken as input in tests below"""

junctions = {K: get_junction_names_from_category(K) for K in junction_categories}
"""Dictionary mapping categories of splice junctions to names of junctions
used in tests below"""

# classification order is repetitive > reference > canonical > untouched

# set up tests as configurations of kwargs
# we'll represent these as 3-tuples, where 1 in a position indicates whether
# or not we'll supply:
#     0. a crossmap
#     1. a reference file
#     2. the --slide_canonical flag
#     3. always value 1 (untouched state)
test_options = list(itertools.product((0, 1), repeat=3))
test_options = [numpy.array(list(X) + [1]) for X in test_options]
base_command_line = " ".join([COMBINED_BED, OUTBASE, "--sequence_file %s" % FASTA_FILE])
switches = [
    " --mask_annotation_format BED --mask_annotation_file %s" % MASK_FILE,
    " --ref %s" % REF_FILE, " --slide_canonical", ""
]
"""Command-line switches used for building tests"""

labels = ["repetitive", "shifted_known", "shifted_canonical", "untouched"]
"""Labels for various junctions under different test conditions"""
# Classification preferences for each set of query junctions
#
# 1 represents the ability to be classified in a category denoted by position,
#
#     0. repetitive
#     1. reference
#     2. canonical
#     3. untouched
#
# These positions correspond to the test_option positions above in the sense
# that if both test_option[position] and classification_pref[k][position] are 1,
# then junctions in group k can be classified with the label corresponding
# to that position. In practice, the leftmost position in which both bits are 1
# should be chosen by the script
classification_prefs = {
    "known_juncs_non_crossmap"    : (0, 1, 1, 1),
    "known_juncs_crossmap"        : (1, 1, 1, 1),
    "to_slide_known_non_crossmap" : (0, 1, 1, 1),
    "to_slide_known_crossmap"     : (1, 1, 1, 1),
    "noncan_no_ref"               : (0, 0, 1, 1),
    "expected_untouched"          : (0, 0, 0, 1),
} # yapf: disable
classification_prefs = {K: numpy.array(V) for K, V in classification_prefs.items()}
"""Classification preferences for each category of query junction"""

output_files = {K: OUTBASE + ("_%s.bed" % K) for K in labels}
"""Output files written by each test"""

# tests will be defined as tuples of command-line arguments,
# and dictionaries mapping classifications to splice junctions that
# should fall into that category for a given run
tests = []
for my_opts in test_options:
    # build command-line arguments
    stmp = base_command_line + " ".join([switches[X] for X in my_opts.nonzero()[0]])
    stmp = stmp % (my_opts[0], my_opts[1], my_opts[2])
    result_dict = {}

    # determine results
    for k, v in classification_prefs.items():
        label = labels[(v & my_opts).argmax()]
        try:
            result_dict[label] |= junctions[k]
        except KeyError:
            result_dict[label] = copy.deepcopy(junctions[k])

    tests.append((my_opts, stmp, result_dict))

#===============================================================================
# INDEX: tests
#===============================================================================


def compare_sets(label, found, expected, crossmap, ref, slide):
    message = "Failed test %s.\n\tCrossmap: %s\n\tRef: %s\n\tslide: %s\nItems expected but not found: %s\nVice versa: %s\n" % (
        label, crossmap, ref, slide, expected - found, found - expected
    )
    assert_set_equal(expected, found, message)


@attr(test="functional")
def do_test():
    for tup, argstr, expected_results in tests:
        test_info["test_method"](shlex.split(argstr))

        for label, filename in output_files.items():
            found = get_junction_names_from_file(filename % (tup[0], tup[1], tup[2]))
            expected = expected_results.get(label, set([]))
            yield compare_sets, label, found, expected, tup[0], tup[1], tup[2]


def tearDownModule():
    """Remove test dataset files after unit tests are complete"""
    if test_info["temp_file_path"] != "":
        shutil.rmtree(test_info["temp_file_path"])

    cleanup_resources()



cdef class CenterMapFactory:
    cdef unsigned int nibble

cdef class FivePrimeMapFactory:
    cdef int offset

cdef class ThreePrimeMapFactory:
    cdef int offset

cdef class VariableFivePrimeMapFactory:
    cdef int [10000] forward_offsets
    cdef int [10000] reverse_offsets

cdef class StratifiedVariableFivePrimeMapFactory(VariableFivePrimeMapFactory):
    cdef int min_length, max_length, _numlengths

cdef class SizeFilterFactory:
    cdef int min_, max_



#!/usr/bin/env python
"""|BigBedReader|, a parser for `BigBed`_ files.

.. contents::
   :local:


Summary
-------

In contrast to `BED`_, `GTF2`_, and `GFF3`_ files, `BigBed`_ files are binary,
indexed, and randomly-accessible. This means:

    - |BigBedReader| can be used to iterate over records, like a reader, **or**
      to fetch records that cover a region of interest, in the manner of a
      |GenomeHash|

    - `BigBed`_ use less memory, because their records don't need to be loaded
      into memory to be parsed or accessed.

    - Indexes `BigBed`_ files can be searched for matching records


Module Contents
---------------

.. autosummary::

   BigBedReader
   BigBedIterator


Examples
--------
Iterate over all features in a `BigBed`_ file::

    >>> my_reader = BigBedReader("some_file.bb", return_type=Transcript)
    >>> for feature in my_reader:
    >>>    pass # do something with each Transcript

`BigBed`_ files can be accessed as dictionaries. To find features overlapping a
region of interest::

    >>> roi = GenomicSegment("chrI", 0, 100000, "+")
    >>> overlapping_features = my_reader[roi]
    >>> list(overlapping_features)
    [ list of SegmentChains/Transcripts ]

Find features that match keyword(s) in a certain field::

    >>> # which fields are indexed and searchable?
    >>> my_reader.indexed_fields
    ['name', 'gene_id']

    >>> # find all entries whose 'gene_id' matches 'nanos'
    >>> list(bb.search('gene_id', 'nanos'))
    [ list of matching SegmentChains/Transcripts ]



See also
--------
`Kent2010 <http://dx.doi.org/10.1093/bioinformatics/btq351>`_
    Description of BigBed and BigWig formats. Especially see supplemental data.

`UCSC file format FAQ <http://genome.ucsc.edu/FAQ/FAQformat.html>`_
    Descriptions of BED, GTF2, GFF3 and other text-based formats.
"""
import itertools
import struct
import sys
import warnings
import zlib
from collections import OrderedDict, Iterable
from plastid.genomics.roitools import GenomicSegment, SegmentChain, add_three_for_stop_codon
from plastid.readers.autosql import AutoSqlDeclaration
from plastid.util.io.binary import BinaryParserFactory, find_null_bytes
from plastid.util.unique_fifo import UniqueFIFO
from plastid.util.services.mini2to3 import ifilter, safe_bytes, safe_str
from plastid.util.services.decorators import skipdoc, deprecated
from plastid.util.services.exceptions import MalformedFileError, FileFormatWarning
from plastid.readers.autosql import AutoSqlDeclaration

from plastid.readers.bbifile cimport (
    bbiFile,
    bits32,
    bits64,
    lm,
    lmInit,
    lmCleanup,
    freeMem,
    _BBI_Reader,
    get_lm,
)

from plastid.genomics.roitools cimport GenomicSegment, SegmentChain
from plastid.genomics.c_common cimport (
    strand_to_str,
    str_to_strand,
    Strand,
    forward_strand,
    reverse_strand,
    unstranded,
    error_strand,
    _GeneratorWrapper,
)

from cpython.mem cimport PyMem_Malloc, PyMem_Free


#===============================================================================
# INDEX: BigBedReader
#===============================================================================

@skipdoc
class _FromBED_StrAdaptor(object):
    """Adaptor class to return strings from |BigBedReaders|.  Is internally
    called by BigBedReader when `return_type` is set to :class:`str`, because
    :class:`str` does not implement a ``from_bed`` method::

        >>> reader = BigBedReader(some_file, return_type=str)
    """
    @staticmethod
    def from_bed(inp):
        """Dummy method. Returns strings as themselves, instead of parsing a
        `BED`_ line

        Parameters
        ----------
        inp : str
            line of `BED`_ formatted text

        Returns
        -------
        str
            `inp`
        """
        return inp

cdef class BigBedReader(_BBI_Reader):
    """
    BigBedReader(filename, return_type = SegmentChain, add_three_for_stop = False, maxmem = 0)

    Reader for `BigBed`_ files. This class is useful for both iteration over
    genomic features one-by-one (like a reader), as well as random access to
    genomic features that overlap a region of interest (like a |GenomeHash|).

    Examples
    --------
    Iterate over all features in a `BigBed`_ file::

        >>> my_reader = BigBedReader("some_file.bb")
        >>> for feature in my_reader:
        >>>    pass # do something with each feature

    `BigBed`_ files can be accessed as dictionaries. To find features
    overlapping a region of interest::

        >>> roi = GenomicSegment("chrI", 0, 100000, "+")
        >>> for feature in my_reader[roi]:
        >>>     pass # do something with that feature

    Find features overlapping a genomic region of interest `roi`, on either
    strand::

        >>> for feature in my_reader.get(roi, stranded=False):
        >>>     pass # do something with that feature


    Parameters
    ----------
    filename : str
        Path to `BigBed`_ file

    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default:
        |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If set to
        `True`, three nucleotides will be added to the threeprime end of each
        CDS annotation, **UNLESS** the annotated transcript contains explicit
        stop_codon feature. (Default: `False`)

    maxmem : float, optional
        Maximum desired memory footprint for C objects, in megabytes.
        May be temporarily exceeded if large queries are requested. Does not
        include memory footprint of Python objects. (Default: 0, no limit)


    Attributes
    ----------
    extension_fields : OrderedDict
        Dictionary mapping custom field names to their descriptions, if any
        custom fields are present

    extension_types : OrderedDict
        Dictionary mapping custom field names to objects that parse their types
        from strings

    filename : str
        Path to open BigBed file

    num_records : int
        Number of features in file

    num_chroms : int
        Number of contigs or chromosomes in file

    chroms : dict
        Dictionary mapping chromosome names to sizes

    return_type : class implementing a :py:meth:`from_bed` method, or str
        Class of object to return (Default: |SegmentChain|)
    """
    def __init__(self,
                 filename,
                 return_type=None,
                 add_three_for_stop=False,
                 **kwargs
                 ):
        """
        Parameters
        ----------
        filename : str
            Path to `BigBed`_ file

        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default:
            |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations.
            If set to `True`, three nucleotides will be added to the threeprime
            end of each CDS annotation, **UNLESS** the annotated transcript
            contains explicit stop_codon feature. (Default: `False`)

        maxmem : float
            Maximum desired memory footprint for C objects, in megabytes.  May
            be temporarily exceeded if large queries are requested.  Does not
            include memory footprint of Python objects.  (Default: 0, no limit)
        """
        cdef:
            str autosql

        self._bbifile = bigBedFileOpen(safe_bytes(filename))

        self.total_fields         = self._bbifile.fieldCount
        self.bed_fields           = self._bbifile.definedFieldCount
        self.num_extension_fields = self.total_fields - self.bed_fields
        self.extension_fields     = OrderedDict()
        self.extension_types      = OrderedDict()

        if self.num_extension_fields > 0:
            autosql = self._get_autosql_str()
            try:
                asql_parser            = AutoSqlDeclaration(autosql)
                self.extension_fields  = OrderedDict(
                    list(asql_parser.field_comments.items())[-self.num_extension_fields:]
                )

                for fieldname in self.extension_fields:
                    self.extension_types[fieldname] = asql_parser.field_formatters[fieldname]

            except AttributeError:
                warnings.warn(
                    "Could not find or could not parse autoSql declaration in BigBed file '%s': %s"
                    % (self.filename, autosql), FileFormatWarning
                )
                for x in range(self.num_extension_fields):
                    fieldname = "custom_%s" % x
                    self.extension_fields[fieldname] = "no description"
                    self.extension_types[ fieldname] = str

        if return_type == None:
            self.return_type = SegmentChain
        elif return_type == str:
            self.return_type = _FromBED_StrAdaptor
        else:
            self.return_type = return_type

        self.add_three_for_stop = add_three_for_stop

    property num_chroms:
        """Number of chromosomes in the `BigBed`_ file"""
        def __get__(self):
            return len(self.c_chroms())

    property num_records:
        """Number of features in file"""
        def __get__(self):
            return bigBedItemCount(self._bbifile)

    property bed_fields:
        """Number of standard `BED`_ format columns included in file"""
        def __get__(self):
            return self._bbifile.definedFieldCount

    property extension_fields:
        """Dictionary of names and types extra fields included in `BigWig`_/`BigBed`_ file"""
        def __get__(self):
            return self.extension_fields

    property return_type:
        """Return type of reader"""
        def __get__(self):
            if self.return_type == _FromBED_StrAdaptor:
                return str
            else:
                return self.return_type

    property indexed_fields:
        """Names of indexed fields in BigBed file. These are searchable by `self.search`"""
        def __get__(self):
            cdef:
                slName *orig_name       = bigBedListExtraIndexes(self._bbifile)
                slName *extension_names = orig_name
                list    names           = []

            if orig_name != NULL:
                names.append(safe_str(extension_names.name))
                while extension_names.next != NULL:
                    extension_names = extension_names.next
                    names.append(safe_str(extension_names.name))

                slFreeList(orig_name)

            return names

    def __str__(self):
        return "<%s records=%s chroms=%s>" % (self.__class__.__name__, self.num_records, self.num_chroms)

    def __repr__(self):
        return str(self)

    def _get_autosql_str(self):
        """Fetch `autoSql`_ field definition string from `BigBed`_ file, if present

        Returns
        -------
        str
            autoSql-formatted string, or "" if no autoSql definition present
        """
        cdef:
            char *  c_asql = bigBedAutoSqlText(self._bbifile)
            str     p_asql

        try:
            if c_asql is NULL:
                p_asql = ""
            else:
                p_asql = safe_str(c_asql)
        finally:
            freeMem(c_asql)

        return p_asql

    cdef list _bigbedinterval_to_bedtext(self, bigBedInterval * iv, Strand strand = unstranded):
        """Convert `bigBedIntervals` to lists of BED-formatted text entries.

        Parameters
        ----------
        iv : *bigBedInterval
            Pointer to first bigBedInterval

        strand : Strand, optional
            Strand that must be overlapped for features to be returned
            (Default: `unstranded`, return all entries)


        Returns
        -------
        list
            List of BED-formatted text entries
        """
        cdef:
            str bed_row, rest
            list ltmp     = []
            dict chromids = self._chromids

        if chromids is None:
            self._define_chroms()
            chromids = self._chromids

        while iv != NULL:
            bed_row = "\t".join("%s" % X for X in [chromids[iv.chromId], iv.start, iv.end])

            if self.total_fields > 3: # if iv.rest is populated
                rest = safe_str(iv.rest)

                # if strand info is present and we care about it
                if self.bed_fields >= 6:
                    items = rest.split("\t")
                    ivstrand = str_to_strand(items[2])

                    # no overlap - restart loop at next iv
                    if ivstrand & strand == 0:
                        iv = iv.next
                        continue

                bed_row = "%s\t%s" % (bed_row, rest)

            ltmp.append(bed_row)
            iv = iv.next

        return ltmp

    def search(self, field_name, *values):
        """Search indexed fields in the `BigBed`_ file for records matching
        `value` See `self.indexed_fields` for names of indexed fields and
        `self.extension_fields` for descriptions of extension fields.

        Parameters
        ----------
        field_name : str
            Name of field to search

        *values : one or more str
            Value(s) to match. If multiple are given, records matching any
            value will be returned.


        Yields
        ------
        object
            `self.return_type` of matching record in the `BigBed`_ file


        Examples
        --------
        Find all entries matching a given gene ID::

            # open file
            >>> bb = BigBedFile("some_file.bb")

            # which fields are searchable?
            >>> bb.indexed_fields
            ['name', 'gene_id']

            # find all entries whose 'gene_id' matches 'nanos'
            >>> bb.search('gene_id', 'nanos')
            [ list of matching segmentchains ]

            # find all entries whose 'gene_id' matches 'nanos' or 'oskar'
            >>> bb.search('gene_id', 'nanos', 'oskar')
            [ list of matching segmentchains ]


        Raises
        ------
        IndexError
            If field `field_name` is not indexed
        """
        cdef:
            int              fieldIdx
            int            * idx = &fieldIdx
            bptFile        * bpt
            bigBedInterval * iv
            lm             * buf = self._get_lm()
            list             ltmp
            object           outfunc   = self.return_type.from_bed
            list             etypes    = list(self.extension_types.items())

            str              stmp
            bytes            val
            char**           vals
            int              n, num_vals

        if field_name not in self.indexed_fields:
            raise KeyError("BigBed file '%s' has no index named '%s'" % (self.filename, field_name))
        else:
            bpt = bigBedOpenExtraIndex(self._bbifile, safe_bytes(field_name), idx)

        if len(values) == 1:
            val  = safe_bytes(values[0])
            iv   = bigBedNameQuery(self._bbifile, bpt, fieldIdx, val, buf)
        else:
            num_vals = len(values)
            vals     = <char**> PyMem_Malloc(num_vals*sizeof(char*))
            if not vals:
                raise MemoryError(
                    "BigBedReader.search(%s, %s): could not allocate memory"
                    % (field_name, ",".join(values))
                )

            for n, stmp in enumerate(values):
                val = safe_bytes(stmp)
                vals[n] = val

            iv  = bigBedMultiNameQuery(self._bbifile, bpt, fieldIdx, vals, num_vals, buf)
            PyMem_Free(vals)

        ltmp = self._bigbedinterval_to_bedtext(iv)
        bptFileDetach(&bpt)

        if self.add_three_for_stop == True:
            return _GeneratorWrapper(
                        (add_three_for_stop_codon(outfunc(X, extra_columns=etypes)) for X in ltmp),
                        "BigBed entries"
                    )
        else:
            return _GeneratorWrapper(
                (outfunc(X, extra_columns=etypes) for X in ltmp),
                "BigBed entries"
            )

    def get(self, roi, bint stranded=True, bint check_unique=True):
        """Iterate over features that share genomic positions with a region of
        interest

        Note
        ----
        ``reader.get(roi)`` is an alternative syntax to ``reader[roi]``. It's
        only useful if setting `stranded` or `check_unique` to `False`.


        Parameters
        ----------
        roi : |SegmentChain| or |GenomicSegment|
            Query feature representing region of interest

        stranded : bool, optional
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands. (Default: `True`)

        check_unique: bool, optional
            if `True`, assure that all results in generator are unique.
            (Default: `True`)


        Yields
        ------
        object
            `self.return_type` of each record in the `BigBed`_ file


        Raises
        ------
        TypeError
            if `other` is not a |GenomicSegment| or |SegmentChain|
        """
        cdef:
            SegmentChain chain

        if isinstance(roi, SegmentChain):
            chain = roi
        elif isinstance(roi, GenomicSegment):
            chain = SegmentChain(roi)
        else:
            raise TypeError("BigBedReader.get(): Query interval must be a GenomicSegment or SegmentChain")

        return self._c_get(chain, stranded, check_unique=check_unique)

    # TODO: direct C/Cython route to SegmentChain.from_bed
    # NB- no cache layer, which we  had in pure Python implementation (below)
    # will this be fast enough for repeated queries over the same region?
    # need to test
    cdef _GeneratorWrapper _c_get(
        self,
        SegmentChain chain,
        bint stranded=True,
        bint check_unique=True,
        lm *my_lm = NULL
        ):
        """c-layer implementation of :meth:`BigBedReader.get`

        Parameters
        ----------
        roi : |SegmentChain| or |GenomicSegment|
            Query feature representing region of interest

        stranded : bool, optional
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands. (Default: `True`)

        check_unique : bool, optional
            If `True`, assure that all results in generator are unique.
            (Default: `True`)

        my_lm : lm, optional
            If not NULL, use this pool of local memory instead of the
            |BigBedReader|'s.  (Default: NULL)


        Yields
        ------
        object
            `self.return_type` of each record in the `BigBed`_ file
        """
        cdef:
            bigBedInterval * iv
            list             ltmp      = []
            GenomicSegment   span      = chain.spanning_segment
            GenomicSegment   roi
            str              chrom     = span.chrom
            Strand           strand    = unstranded
            lm*              buf #       = self._get_lm()
            Strand           ivstrand
            object           outfunc   = self.return_type.from_bed
            list             etypes    = list(self.extension_types.items())

        if my_lm != NULL:
            buf = my_lm
        else:
            buf = self._get_lm()

        if stranded is True:
            strand = span.c_strand

        for roi in chain:
            iv = bigBedIntervalQuery(self._bbifile,
                                     safe_bytes(span.chrom),
                                     roi.start,
                                     roi.end,
                                     0,
                                     buf)
            ltmp.extend(self._bigbedinterval_to_bedtext(iv, strand=strand))

        # filter for uniqueness
        if check_unique == True:
            ltmp = sorted(set(ltmp))

        if self.add_three_for_stop == True:
            return _GeneratorWrapper((
                    add_three_for_stop_codon(outfunc(X, extra_columns=etypes)) for X in ltmp
                    ),
                "BigBed entries"
                )
        else:
            return _GeneratorWrapper((outfunc(X, extra_columns=etypes) for X in ltmp), "BigBed entries")

    def __getitem__(self, roi):
        """Iterate over features that share genomic positions with a region of
        interest, on same strand.  Unstranded features, if present, are
        considered to overlap both `rois` on any strand.


        Parameters
        ----------
        roi : |SegmentChain| or |GenomicSegment|
            Query feature representing region of interest

        Yields
        ------
        object
            `self.return_type` of each record in the `BigBed`_ file


        Raises
        ------
        TypeError
            if `other` is not a |GenomicSegment| or |SegmentChain|
        """
        return self.get(roi, stranded=True)

    def __iter__(self):
        """Iterate over all features in `BigBed`_ file

        Yields
        ------
        object
            Object of ``self.return_type``, |SegmentChain| or one of its
            subclasses
        """
        return _GeneratorWrapper(BigBedIterator(self, maxmem=self._maxmem), "BigBed records")


# can't be cdef'ed or cpdef'ed due to yield
#
# This would probably be faster if we iterated through the cirTree leaf nodes
# directly, since they overlap chromosome boundaries. But, as long as the
# number of chromosomes/contigs in the file is low, this shouldn't be too big
# a deal.
#
# But, cirTree.h, crTree.h, and BigBed.h don't give a convenient way to get
# all the data from a node, so doing so would result in lots of reimplementation.
def BigBedIterator(BigBedReader reader, maxmem=0):
    """BigBedIterator(reader, maxmem = 0)

    Iterate over records in the `BigBed`_ file, sorted lexically by chromosome
    and position.

    Parameters
    ----------
    reader : |BigBedReader|
        Reader to iterate over

    maxmem : float
        Maximum desired memory footprint for C objects, in megabytes.  May be
        temporarily exceeded if large queries are requested.  Does not include
        memory footprint of Python objects.  (Default: 0, no limit)

    Yields
    ------
    object
        reader.return_type of `BED`_ record

    Raises
    ------
    MemoryError
        If memory cannot be allocated
    """
    cdef:
        list           chromsizes  = sorted(reader.c_chroms().items(), key = lambda x: x[0].lower())
        long           chromlength
        str            chrom
        lm             *buf = lmInit(0)

    if not buf:
        raise MemoryError("BigBedIterator: could not allocate local memory")

    for chrom, chromlength in chromsizes:
        query = SegmentChain(GenomicSegment(chrom, 0, chromlength, "."))
        buf = get_lm(my_lm=buf, maxmem=maxmem)
        if not buf:
            raise MemoryError("BigBedIterator: could not allocate local memory")

        for n, roi in enumerate(reader._c_get(query, stranded=False, check_unique=False, my_lm=buf)):
            yield roi

    lmCleanup(&buf)


#===============================================================================
# INDEX: Factories for various record formats
#===============================================================================

HeaderFactory = BinaryParserFactory(
    "BigBedHeader",
    "IHH3QHHQQIQ",
    [
        "magic",
        "version",
        "zoom_levels",
        "bplus_tree_offset",
        "full_data_offset",
        "r_tree_offset",
        "field_count",
        "bed_field_count",
        "autosql_offset",
        "total_summary_offset",
        "uncompressed_buffer_size",
        "reserved",
    ],
)
"""Parse headers for BigWig and `BigBed`_ files"""

ZoomHeaderFactory = BinaryParserFactory(
    "ZoomHeader",
    "2H2Q",
    [
        "reduction_level",
        "reserved",
        "data_offset",
        "index_offset",
    ],
)
"""Parse Zoomlevel headers in BigWig files"""


TotalSummaryFactory = BinaryParserFactory(
    "TotalSummary",
    "5Q",
    [
        "bases_covered",
        "min_val",
        "max_val",
        "sum_data",
        "sum_squares",
    ],
)
"""Parse 'Total Summary' tables in BigWig and `BigBed`_ files"""

BPlusTreeHeaderFactory = BinaryParserFactory(
    "BPlusTreeHeader",
    "4I2Q",
    [
        "magic",
        "block_size",
        "key_size",
        "val_size",
        "num_chroms",
        "reserved",
    ],
)
"""Parse headers of |BPlusTree|"""

BPlusTreeNodeFormatFactory  = BinaryParserFactory(
    "BPlusTreeNodeFormat",
    "?BH",
    ["is_leaf", "reserved", "child_count"],
)
"""Determine format of BPlus Tree nodes in a datablock of a `BigBed`_ or BigWig file"""

RTreeHeaderFactory = BinaryParserFactory(
    "RTreeHeader",
    "IIQ4IQII",
    [
        "magic",
        "block_size",
        "num_nodes",
        "start_chrom_id",
        "start_base",
        "end_chrom_id",
        "end_base",
        "end_file_offset",
        "items_per_leaf",
        "reserved",
    ]
)
"""Parse headers of |RTree|"""

RTreeNodeFormatFactory = BinaryParserFactory(
    "RTreeNodeFormat",
    "BBH",
    ["is_leaf", "reserved", "count"],
    )
"""Determine format of R Tree nodes in a data block of a `BigBed`_ or BigWig file"""


RTreeLeafFactory = BinaryParserFactory(
    "RTreeLeaf",
    "4I2Q",
    [
        "start_chrom_id",
        "start_base",
        "end_chrom_id",
        "end_base",
        "data_offset",
        "data_size",
    ],
)
"""Parse leaves of |RTree|"""

RTreeNonLeafFactory = BinaryParserFactory(
    "RTreeNonLeaf",
    "4IQ",
    [
        "start_chrom_id",
        "start_base",
        "end_chrom_id",
        "end_base",
        "child_data_offset",
    ],
)
"""Parse non-leaf nodes of |RTree|"""



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.readers.bigbed`

Notes
-----
Several of these tests are tested against |GenomeHash|, and so will fail if 
|GenomeHash| is malfunctioning
"""

import unittest
import copy
import warnings
from random import shuffle
from pkg_resources import resource_filename, cleanup_resources
from nose.plugins.attrib import attr
from nose.tools import assert_almost_equal

from collections import OrderedDict
from plastid.genomics.roitools import SegmentChain, GenomicSegment, Transcript
from plastid.genomics.genome_hash import GenomeHash
from plastid.readers.bed import BED_Reader
from plastid.readers.bigbed import BigBedReader

warnings.simplefilter("ignore", DeprecationWarning)

#===============================================================================
# INDEX: helper functions
#===============================================================================


def tearDownModule():
    """Remove test dataset files after unit tests are complete"""
    cleanup_resources()


def transcript_identical(ivc1, ivc2):
    """Test for identity between positions of two Transcripts"""
    position_test = ivc1.get_position_set() == ivc2.get_position_set()
    strand_test = ivc1.spanning_segment.strand == ivc2.spanning_segment.strand
    chrom_test = ivc1.spanning_segment.chrom == ivc2.spanning_segment.chrom

    start_test = (ivc1.cds_start is None and ivc2.cds_start is None) or\
                 (ivc1.cds_start == ivc2.cds_start)
    end_test   = (ivc1.cds_end is None and ivc2.cds_end is None) or\
                 (ivc1.cds_end == ivc2.cds_end)

    return position_test & strand_test & chrom_test & start_test & end_test


#===============================================================================
# INDEX: test suites
#===============================================================================


@attr(test="unit")
class test_BigBedReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cols = [3, 4, 5, 6, 8, 9, 12]
        cls.bedfiles = {}
        cls.bbfiles = {}
        for col in cls.cols:
            cls.bedfiles[col] = resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed%s.bed" % col
            )
            cls.bbfiles[col] = resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed%s.bb" % col
            )

        cls.chrom_sizes = {}
        with open(resource_filename("plastid", "test/data/annotations/sacCer3.sizes")) as fh:
            for line in fh:
                chrom, size = line.strip().split("\t")
                cls.chrom_sizes[chrom] = int(size)

        cls.bbs = {K: BigBedReader(cls.bbfiles[K], return_type=Transcript) for K in cls.cols}

        # comparisons against genome hash
        cls.binsize = 10000
        with open(cls.bedfiles[12]) as fh:
            transcripts = list(BED_Reader(fh, return_type=Transcript))

        cls.tx_dict = {}
        cls.cds_dict = {}
        cls.as_cds_dict = {}
        for tx in transcripts:
            txid = tx.get_name()
            cls.tx_dict[txid] = tx
            cds_ivc = tx.get_cds()
            cds_ivc.attr["ID"] = txid
            if cds_ivc.length > 0:
                cls.cds_dict[txid] = tx.get_cds()
                cls.as_cds_dict[txid] = tx.get_cds().get_antisense()
                cls.as_cds_dict[txid].attr["ID"] = txid

        cls.tx_hash = GenomeHash(cls.tx_dict, do_copy=False, binsize=cls.binsize)
        cls.cds_hash = GenomeHash(cls.cds_dict, do_copy=False, binsize=cls.binsize)
        cls.as_cds_hash = GenomeHash(cls.as_cds_dict, do_copy=False, binsize=cls.binsize)

        cls.shuffled_indices = list(range(len(transcripts)))
        shuffle(cls.shuffled_indices)

        cls.flybbfile = resource_filename(
            "plastid", "test/data/annotations/dmel-all-no-analysis-r5.54.bb"
        )
        cls.flybedfile = resource_filename(
            "plastid", "test/data/annotations/dmel-all-no-analysis-r5.54.bed"
        )

        # BigBed files with and without extra columns, with and without autoSql descriptions
        cls.bb_bonuscols = {
            "bb4as":
            resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed4plus_bonus_as.bb"
            ),
            "bb12as":
            resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed12plus_bonus_as.bb"
            ),
            "bb4no_as":
            resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed4plus_bonus_no_as.bb"
            ),
            "bb12no_as":
            resource_filename(
                "plastid", "test/data/annotations/100transcripts_bed12plus_bonus_no_as.bb"
            ),
        }
        cls.bonus_col_file = resource_filename(
            "plastid", "test/data/annotations/bonus_bed_columns.txt"
        )

        # BigBed file with indexes
        cls.bb_indexed = resource_filename("plastid", "test/data/annotations/dmel-bonus-cols.bb")

    def test_count_records(self):
        for _, my_reader in self.bbs.items():
            # make sure we have all records
            self.assertEqual(my_reader.num_records, 100)

    def test_num_chroms(self):
        for _, my_reader in self.bbs.items():
            self.assertEqual(my_reader.num_chroms, 17)

    def test_chrom_sizes(self):
        for _, my_reader in self.bbs.items():
            for k, v in self.chrom_sizes.items():
                self.assertEqual(my_reader.chroms[k], v)

    def test_iter_same_as_bed_reader_various_columns(self):
        # implicitly tests iterate_over_chunk over all bed files, too
        # this tests BigBed equality with various ranges of columns
        # and various custom columns
        for col in self.cols:
            bigbed = self.bbs[col]
            with open(self.bedfiles[col]) as fh:
                bed = BED_Reader(fh, return_type=Transcript)

                for n, (tx1, tx2) in enumerate(zip(bed, bigbed)):
                    msg = "Transcript mismatch in BigBed file at record %s. Expected '%s'. Got '%s'." % (
                        n, tx1, tx2
                    )
                    self.assertTrue(transcript_identical(tx1, tx2), msg)

            self.assertEqual(n, 100 - 1)

    def test_iter_same_as_bed_reader_flydata(self):
        # test more complex transcript models
        # we cast them to lists, sadly, because Python's lexical chromosome sorting
        # differs from unix command-line sort; so even though the records are
        # in the same order in both files, they are returned with different sorts
        flybb = BigBedReader(self.flybbfile, return_type=Transcript)
        with open(self.flybedfile) as fh:
            flybed = BED_Reader(fh, return_type=Transcript)
            for n, (tx1, tx2) in enumerate(zip(flybed, flybb)):
                msg = "Transcript mismatch in BigBed file at record %s. Expected '%s'. Got '%s'." % (
                    n, tx1, tx2
                )
                self.assertTrue(transcript_identical(tx1, tx2), msg)

        self.assertEqual(n, 32682 - 1)

    def test_getitem_stranded(self):
        """Test fetching of overlapping features, minding strand
        
        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        
        2.  Make sure each feature cannot fetch its own antisense subregion
        
        3.  Make sure each features fetches exactly the same features as a GenomeHash
        """
        # make sure we can fetch each transcript's own CDS
        bb = self.bbs[12]
        u = 0
        for txid, cds in list(self.cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=True)
            bb_ol_features = bb[cds]

            self.assertIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s failed to fetch its own CDS on correct strand" % txid
            )

            # make sure bb fetch matches GenomeHash fetch
            self.assertSetEqual(
                set([str(X) for X in gh_ol_features]), set([str(X) for X in bb_ol_features])
            )

            u += 1

        self.assertGreater(u, 0)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, cds in list(self.as_cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=True)
            bb_ol_features = bb[cds]
            self.assertNotIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )
            self.assertSetEqual(
                set([str(X) for X in gh_ol_features]), set([str(X) for X in bb_ol_features])
            )

    def test_get_stranded(self):
        """Test fetching of overlapping features, minding strand
        
        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        
        2.  Make sure each feature cannot fetch its own antisense subregion
        
        3.  Make sure each features fetches exactly the same features as a GenomeHash
        """
        # make sure we can fetch each transcript's own CDS
        bb = self.bbs[12]
        u = 0
        for txid, cds in list(self.cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=True)
            bb_ol_features = bb.get(cds, stranded=True)

            self.assertIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s failed to fetch its own CDS on correct strand" % txid
            )

            # make sure bb fetch matches GenomeHash fetch
            self.assertSetEqual(
                set([str(X) for X in gh_ol_features]), set([str(X) for X in bb_ol_features])
            )

            u += 1

        self.assertGreater(u, 0)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, cds in list(self.as_cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=True)
            bb_ol_features = bb[cds]
            self.assertNotIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )
            self.assertSetEqual(
                set([str(X) for X in gh_ol_features]), set([str(X) for X in bb_ol_features])
            )

    def test_get_unstranded(self):
        """Test fetching of overlapping features, disregarding strand
        
        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        
        2.  Make sure each feature can fetch its own antisense subregion
        
        3.  Make sure each features fetches exactly the same features as a GenomeHash
        """
        # make sure we can fetch each transcript's from its own CDS on same strand
        bb = self.bbs[12]
        u = 0
        for txid, cds in list(self.cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=False)
            bb_ol_features = bb.get(cds, stranded=False)
            self.assertIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s failed to fetch its own CDS on same strand" % txid
            )

            # make sure bb fetch matches GenomeHash fetch
            self.assertSetEqual(
                set([str(X) + X.get_name() for X in gh_ol_features]),
                set([str(X) + X.get_name() for X in bb_ol_features])
            )

            u += 1

        self.assertGreater(u, 0)

        # make sure we can fetch each transcript's from its own antisense CDS
        # on opposite strand
        for txid, cds in list(self.as_cds_dict.items())[:100]:
            gh_ol_features = self.tx_hash.get_overlapping_features(cds, stranded=False)
            bb_ol_features = bb.get(cds, stranded=False)
            self.assertIn(
                txid, (X.get_name() for X in gh_ol_features),
                msg="%s failed to fetched its own name on opposite strand!" % txid
            )
            s1 = set([str(X) + X.get_name() for X in gh_ol_features])
            s2 = set([str(X) + X.get_name() for X in bb_ol_features])
            self.assertSetEqual(
                s1,
                s2,
                msg="%s failure:\n    Only in first set: %s\n    Only in second set: %s" %
                (txid, s1 - s2, s2 - s1)
            )

    def test_return_type(self):
        bb = self.bbs[12]
        i = iter(bb)
        for _ in range(5):
            self.assertTrue(isinstance(next(i), Transcript))
        ivcbb = BigBedReader(self.bbfiles[12], return_type=SegmentChain)
        i = iter(ivcbb)
        for _ in range(5):
            self.assertTrue(isinstance(next(i), SegmentChain))

    def test_get_autosql_str(self):
        for k in (4, 12):
            bbplus_as = BigBedReader(self.bb_bonuscols["bb%sas" % k])
            with open(resource_filename(
                    "plastid", "test/data/annotations/bed%s_bonus_bed_columns.as" % k)) as fh:
                expected_as = fh.read()

            self.assertEqual(bbplus_as._get_autosql_str(), expected_as)

    def test_get_no_autosql_str(self):
        for k in (4, 12):
            bbplus_noas = BigBedReader(self.bb_bonuscols["bb%sno_as" % k])
            self.assertEqual(bbplus_noas._get_autosql_str(), "")

    def test_custom_columns_names_with_autosql(self):
        expected = OrderedDict(
            [
                ("my_floats", "some float values"),
                ("my_sets", "some set options"),
                ("my_ints", "signed integer values"),
                ("my_strs", "str representation of transcripts"),
                ("my_colors", "r,g,b colors"),
            ]
        )
        for k in (4, 12):
            fn = "bb%sas" % k
            bb = BigBedReader(self.bb_bonuscols[fn])
            self.assertEqual(bb.extension_fields, expected)

    def test_custom_columns_names_without_autosql(self):
        expected = OrderedDict(
            [
                ("custom_0", "no description"),
                ("custom_1", "no description"),
                ("custom_2", "no description"),
                ("custom_3", "no description"),
                ("custom_4", "no description"),
            ]
        )
        for k in (4, 12):
            fn = "bb%sno_as" % k
            bb = BigBedReader(self.bb_bonuscols[fn])
            self.assertEqual(bb.extension_fields, expected)

    def test_custom_columns_retval_type_with_autosql(self):
        values = {
            "my_floats": [],
            "my_sets": [],
            "my_ints": [],
            "my_strs": [],
            "my_colors": [],
        }

        with open(self.bonus_col_file) as bfile:
            for line in bfile:
                items = line.strip("\n").split("\t")
                values["my_floats"].append(float(items[0]))
                if items[1] == "":
                    values["my_sets"].append(set())
                else:
                    values["my_sets"].append(set([X.strip() for X in items[1].split(",")]))
                values["my_ints"].append(int(items[2]))
                values["my_strs"].append(items[3])
                values["my_colors"].append(tuple([int(X) for X in items[4].split(",")]))

        for k in (4, 12):
            fn = "bb%sas" % k
            # ignore a Warning caused by trying to turn the BED color field
            # to an int- this has to deal with the fact that BedToBigBed wants
            # field 9 (itemRgb, typically uint[3]) to be `reserved uint;` in
            # autoSql declarations
            with warnings.catch_warnings():
                #warnings.simplefilter("ignore")
                bb = BigBedReader(self.bb_bonuscols[fn])
                for n, item in enumerate(bb):
                    for key in values:
                        expected = values[key][n]
                        found = item.attr[key]
                        msg = "failed test_custom_columns_retval_type_with_autosql at record %s, key %s. Expected '%s'. Got '%s' " % (
                            n, key, expected, found
                        )
                        if isinstance(expected, float):
                            assert_almost_equal(expected, found, msg)
                        else:
                            self.assertEqual(expected, found, msg)

    def test_custom_columns_retval_type_without_autosql(self):
        values = {"custom_%s" % X: copy.deepcopy([]) for X in range(5)}
        with open(self.bonus_col_file) as bfile:
            for line in bfile:
                items = line.strip("\n").split("\t")
                values["custom_0"].append(items[0])
                values["custom_1"].append(items[1])
                values["custom_2"].append(items[2])
                values["custom_3"].append(items[3])
                values["custom_4"].append(items[4])

        for k in (4, 12):
            fn = "bb%sno_as" % k
            bb = BigBedReader(self.bb_bonuscols[fn])
            for n, item in enumerate(bb):
                for key in values:
                    self.assertEqual(values[key][n], item.attr[key])

    def test_indexed_fields(self):
        reader = BigBedReader(self.bb_indexed)
        self.assertEqual(
            sorted(["gene_id", "name", "name2", "Alias"]), sorted(reader.indexed_fields)
        )

    def test_indexed_fields_no_as_no_index(self):
        reader = BigBedReader(self.bb_bonuscols["bb12no_as"])
        self.assertEqual([], reader.indexed_fields)

    def test_indexed_fields_as_no_index(self):
        reader = BigBedReader(self.bb_bonuscols["bb4as"])
        self.assertEqual([], reader.indexed_fields)

    def test_search_fields_invalid_raises_error(self):
        reader = BigBedReader(self.bb_indexed)
        self.assertRaises(KeyError, reader.search, "garbage_field", "garbage_value")

    def test_search_fields_singlevalue(self):
        reader = BigBedReader(self.bb_indexed)
        found = list(reader.search("name", "should_have_no_match"))
        self.assertEqual([], found)

        found = list(reader.search("name2", "Sam-S-RE"))
        expected = [
            SegmentChain(
                GenomicSegment('2L', 106902, 107000, '+'),
                GenomicSegment('2L', 107764, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RE', 'CG2674-RE']'",
                ID='FBtr0089437',
                name2='Sam-S-RE',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
        ]
        self.assertEqual(expected, found)

        found = list(reader.search("gene_id", "FBgn0005278"))
        expected = [
            SegmentChain(
                GenomicSegment('2L', 106902, 107000, '+'),
                GenomicSegment('2L', 107764, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RE', 'CG2674-RE']'",
                ID='FBtr0089437',
                name2='Sam-S-RE',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107760, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 111337, '+'),
                Alias='na',
                ID='FBtr0308091',
                name2='Sam-S-RK',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='110900',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107760, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111004, 111117, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114210, '+'),
                Alias="'['M(2)21AB-RB', 'CG2674-RB']'",
                ID='FBtr0089428',
                name2='Sam-S-RB',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='112741',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107760, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RA', 'CG2674-RA']'",
                ID='FBtr0089429',
                name2='Sam-S-RA',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107760, 107956, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias='na',
                ID='FBtr0330656',
                name2='Sam-S-RL',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='112781',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107936, 108226, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114210, '+'),
                Alias="'['M(2)21AB-RH', 'CG2674-RH']'",
                ID='FBtr0089432',
                name2='Sam-S-RH',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107936, 108101, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RD', 'CG2674-RD']'",
                ID='FBtr0089430',
                name2='Sam-S-RD',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107936, 108101, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111004, 111117, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RC', 'CG2674-RC']'",
                ID='FBtr0089431',
                name2='Sam-S-RC',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 108088, 108226, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RF', 'CG2674-RF']'",
                ID='FBtr0089433',
                name2='Sam-S-RF',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 108132, 108346, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RI', 'CG2674-RI']'",
                ID='FBtr0089434',
                name2='Sam-S-RI',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 108132, 108226, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111004, 111117, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RJ', 'CG2674-RJ']'",
                ID='FBtr0089435',
                name2='Sam-S-RJ',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 109593, 109793, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111004, 111117, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114210, '+'),
                Alias="'['M(2)21AB-RG', 'CG2674-RG']'",
                ID='FBtr0089436',
                name2='Sam-S-RG',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='109750',
                type='exon'
            ),
        ]
        self.assertEqual(sorted(expected), sorted(found))

    def test_search_fields_multivalue(self):
        reader = BigBedReader(self.bb_indexed)
        found = list(reader.search("name", "should_have_no_match", "should_also_have_no_match"))
        self.assertEqual([], found)
        found = list(reader.search("name2", "Sam-S-RE", "Sam-S-RK"))
        expected = [
            SegmentChain(
                GenomicSegment('2L', 106902, 107000, '+'),
                GenomicSegment('2L', 107764, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 110877, '+'),
                GenomicSegment('2L', 111906, 112019, '+'),
                GenomicSegment('2L', 112689, 113369, '+'),
                GenomicSegment('2L', 113433, 114432, '+'),
                Alias="'['M(2)21AB-RE', 'CG2674-RE']'",
                ID='FBtr0089437',
                name2='Sam-S-RE',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='113542',
                thickstart='108685',
                type='exon'
            ),
            SegmentChain(
                GenomicSegment('2L', 107760, 107838, '+'),
                GenomicSegment('2L', 108587, 108809, '+'),
                GenomicSegment('2L', 110405, 110483, '+'),
                GenomicSegment('2L', 110754, 111337, '+'),
                Alias='na',
                ID='FBtr0308091',
                name2='Sam-S-RK',
                color='#000000',
                gene_id='FBgn0005278',
                score='0.0',
                thickend='110900',
                thickstart='108685',
                type='exon'
            ),
        ]
        self.assertEqual(expected, found)



"""Test suite for :py:mod:`plastid.genomics.seqtools`"""
import re
import sys
from nose.plugins.attrib import attr
from nose.tools import assert_equal, assert_set_equal, assert_true
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from plastid.test.ref_files import REF_FILES
from plastid.genomics.seqtools import seq_to_regex, mutate_seqs, random_seq,\
                                   _TwoBitSeqProxy, TwoBitSeqRecordAdaptor

#===============================================================================
# INDEX: unit tests
#===============================================================================


@attr(test="unit")
def test_seq_to_regex():
    for flags, inp, outp in SEQ_REGEX:
        check = seq_to_regex(inp, flags=flags)
        pattern = check.pattern
        pflags = check.flags
        yield assert_equal, pattern, outp, "seq_to_regex(%s,%s) regex do not match. Expected %s, found %s" % (
            inp, flags, outp, pattern
        )
        yield assert_equal, flags, pflags, "seq_to_regex(%s,%s) flags do not match. Expected %s, found %s" % (
            inp, flags, flags, pflags
        )


@attr(test="unit")
def test_mutate_seqs():
    for inp, nucs, mutations, outp in MUTANTS:
        check = mutate_seqs(inp, nucleotides=nucs, mutations=mutations)
        yield assert_set_equal, check, outp, "mutate_seqs(%s,%s) did not match expectation. Expected %s, found %s" % (
            inp, mutations, outp, check
        )


def check_random_seq(length, nucleotides):
    check = random_seq(length, nucleotides=nucleotides)

    # check length
    assert_equal(
        length, len(check), "random_chr(%s,%s) length is incorrect. Expected %s, found %s" %
        (length, nucleotides, length, len(check))
    )

    # make sure nucleotides used are a subset of those expected
    diff = set(list(check)) - set(nucleotides)
    assert_equal(
        diff, set(),
        "random_chr(%s,%s) found extra nucleotides in result: %s" % (length, nucleotides, diff)
    )


@attr(test="unit")
def test_random_chr():
    for length, nucs in RANDOM_CHR:
        yield check_random_seq, length, nucs


@attr(test="unit")
def testTwoBitSeqProxyFetch():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    seq = g2["chrV"][50:5000]
    assert_true(isinstance(seq, SeqRecord))
    assert_equal(str(g1["chrV"][50:5000].seq), str(seq.seq))


@attr(test="unit")
def testTwoBitSeqProxyLen():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    for k in g1:
        assert_equal(len(g1[k]), len(g2[k]))


@attr(test="unit")
def testTwoBitSeqProxyStr():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    for k in g1:
        assert_equal(str(g1[k].seq), str(g2[k]))


@attr(test="unit")
def testTwoBitSeqProxySeqProp():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    for k in g1:
        assert_equal(str(g1[k].seq), str(g2[k].seq))


@attr(test="unit")
def testTwoBitSeqProxyRevComp():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    for k in g1:
        assert_equal(str(g1[k].reverse_complement().seq), str(g2[k].reverse_complement().seq))


@attr(test="unit")
def testTwoBitSeqRecordAdaptor():
    g1 = SeqIO.to_dict(SeqIO.parse(REF_FILES["yeast_fasta"], "fasta"))
    g2 = TwoBitSeqRecordAdaptor(REF_FILES["yeast_twobit"])

    for k, v in g1.items():
        assert_true(isinstance(g2[k], _TwoBitSeqProxy))


#===============================================================================
# INDEX: test data
#===============================================================================

if sys.version_info > (3, ):
    default_flag = re.UNICODE
else:
    default_flag = 0

SEQ_REGEX = [
    (default_flag, "NNNN", "[ACTGU][ACTGU][ACTGU][ACTGU]"),  # N->ACTGU
    (default_flag, "CCCAGA", "CCCAGA"),                      # No change
    (default_flag, "TCTAGA", "[TU]C[TU]AGA"),                # T->TU
    (default_flag, "UCUAGA", "[TU]C[TU]AGA"),                # U->TU
    (default_flag, "ARCGA", "A[AG]CGA"),                     # R->AG
    (default_flag, "YAGAC", "[CTU]AGAC"),                    # Y->CTU
    (default_flag, "SAGAC", "[GC]AGAC"),                     # S->GC
    (default_flag, "WAGAC", "[ATU]AGAC"),                    # W->AT
    (default_flag, "KAGAC", "[GTU]AGAC"),
    (default_flag, "MAGAC", "[AC]AGAC"),
    (default_flag, "BAGAC", "[CGTU]AGAC"),
    (default_flag, "DAGAC", "[AGTU]AGAC"),
    (default_flag, "HAGAC", "[ACTU]AGAC"),
    (default_flag, "VAGAC", "[ACG]AGAC"),
    (default_flag, "SWANKY", "[GC][ATU]A[ACTGU][GTU][CTU]"),
    (default_flag | re.IGNORECASE, "KAGAC", "[GTU]AGAC"),
    (default_flag | re.DEBUG, "MAGAC", "[AC]AGAC"),
    (default_flag | re.MULTILINE, "BAGAC", "[CGTU]AGAC"),
    (default_flag | re.DEBUG | re.MULTILINE, "DAGAC", "[AGTU]AGAC"),
    (default_flag | re.DOTALL | re.UNICODE, "HAGAC", "[ACTU]AGAC"),
] # yapf: disable
"""Test cases for :py:func:`seq_to_regex`, as tuples of
    1. Regex flags
    2. Input string
    3. Expected regex pattern
"""

MUTANTS = [
    ("A", "NACTG", 1, {"A", "C", "T", "G", "N"}),
    ("C", "NACTG", 1, {"A", "C", "T", "G", "N"}),
    ("T", "NACTG", 1, {"A", "C", "T", "G", "N"}),
    ("G", "NACTG", 1, {"A", "C", "T", "G", "N"}),
    ("N", "NACTG", 1, {"A", "C", "T", "G", "N"}),
    # 2 mutations
    ("A", "NACTG", 2, {"A", "C", "T", "G", "N"}),
    ("AT", "NACTG", 1, {"AT", "AA", "AC", "AG", "AN", "CT", "GT", "TT", "NT"}),
    # 2 mutations
    (
        "AT", "NACTG", 2, {
            "AA", "AC", "AT", "AG", "AN", "CA", "CC", "CT", "CG", "CN", "GA", "GC", "GT", "GG",
            "GN", "TA", "TC", "TT", "TG", "TN", "NA", "NC", "NT", "NG", "NN"
        }
    ),
    (
        "AAA", "NACTG", 1,
        {"AAA", "AAT", "AAC", "AAG", "AAN", "ATA", "ACA", "AGA", "ANA", "TAA", "CAA", "GAA", "NAA"}
    ),
    (
        "AAAA", "NACTG", 1, {
            "AAAA",
            "AAAT",
            "AAAC",
            "AAAG",
            "AAAN",
            "AATA",
            "AACA",
            "AAGA",
            "AANA",
            "ATAA",
            "ACAA",
            "AGAA",
            "ANAA",
            "TAAA",
            "CAAA",
            "GAAA",
            "NAAA",
        }
    ),
    # restricted nucleotides
    ("AAAA", "CT", 1, {
        "AAAA",
        "AAAT",
        "AAAC",
        "AATA",
        "AACA",
        "ATAA",
        "ACAA",
        "TAAA",
        "CAAA",
    }),
    # multiple input sequences
    (
        ["A", "AAA"], "NACTG", 1, {
            "A", "C", "T", "G", "N", "AAA", "AAT", "AAC", "AAG", "AAN", "ATA", "ACA", "AGA", "ANA",
            "TAA", "CAA", "GAA", "NAA"
        }
    )
]
"""Test cases for :py:func:`mutate_seqs` as tuples of:
    1. Input sequence(s)
    2. Permitted nucleotide substitutions
    3. Number of mutations
    4. Expected output
"""

RANDOM_CHR = [
    (1, "ACTG"),
    (100, "AC"),
    (100, "ACTG"),
    (10000, "ACTG"),
]
"""Test cases for :py:func:`random_seq` as tuples of:
    1. Sequence length
    2. Nucleotide composition
"""



#!/usr/bin/env python
"""Tools for reading, writing, analyzing, and manipulating GFF file subtypes
(e.g. `GTF2`_ and `GFF3`_).

.. contents::
   :local:


Summary
-------

Because `GTF2`_/`GFF3`_ files are hierarchically structured -- i.e. a complex
feature can be assembled from several component features; each component
feature having its own record on its own line -- two interfaces for reading
`GTF2`_/`GFF3`_ files are included:

    Assembly of transcripts from exon, CDS, & UTR annotations
        |GTF2_TranscriptAssembler| and |GFF3_TranscriptAssembler| collect
        individual exon and CDS features, and assemble these into |Transcripts|.

        Features are read from `GTF2`_/`GFF3`_ files, grouped by `transcript_id`,
        `Parent`, or `ID` attributes, depending on file type. Assembled
        |Transcripts| are yielded only when their component features have fully
        been collected.

    Low-level parsing of simple features
        |GTF2_Reader| and |GFF3_Reader| read raw features (such as individual
        exons, stop codons, SNPs, et c) from `GTF2`_/`GFF3`_ files. Each line
        is returned as a |SegmentChain|.

Module contents
---------------

.. autosummary::

   GTF2_Reader
   GTF2_TranscriptAssembler
   GFF3_Reader
   GFF3_TranscriptAssembler


Examples
--------

|GTF2_Reader| and |GFF3_Reader| return raw, unmodified features from `GTF2`_ or
`GFF3`_ files -- e.g. exons, coding regions, stop codons -- without assembling
them into transcripts::

    >>> feature_reader = GTF2_Reader("some_file.gtf")
    >>> for feature in reader:
    >>>     print(feature.get_name(),feature.attr["type"],str(feature))
    ('YAL030W_mRNA',  'exon',        'chrI:87262-87387(+)')
    ('YAL030W_mRNA',  'exon',        'chrI:87500-87857(+)')
    ('YAL030W_mRNA',  'CDS',         'chrI:87285-87387(+)')
    ('YAL030W_mRNA',  'CDS',         'chrI:87500-87749(+)')
    ('YAL030W_mRNA',  'start_codon', 'chrI:87285-87288(+)')
    ('YAL030W_mRNA',  'stop_codon',  'chrI:87749-87752(+)')
    ('YBL092W_mRNA',  'exon',        'chrII:45643-45644(+)')
    ('YBL092W_mRNA',  'exon',        'chrII:45977-46440(+)')
    ('YBL092W_mRNA',  'CDS',         'chrII:45977-46367(+)')
    ('YBL092W_mRNA',  'start_codon', 'chrII:45977-45980(+)')
    [rest of output omitted]

In contrast, |GTF2_TranscriptAssembler| and |GFF3_TranscriptAssembler| reconstruct
transcripts from their components, based upon their `transcript_id`, `ID`, or
`Parent` attributes. Note how all features are of type `mRNA`, and how some
contain multiple exons (coordinates separated by `'^'`)::

    >>> transcript_reader = GTF2_TranscriptAssembler("some_file.gtf")
    >>> for transcript in reader:
    >>>     print(transcript.get_name(),transcript.attr["type"],str(transcript))
    ('YAL030W_mRNA',   'mRNA',  'chrI:87262-87387^87500-87857(+)')
    ('YBL092W_mRNA',   'mRNA',  'chrII:45643-45644^45977-46440(+)')
    ('YBL057C_mRNA',   'mRNA',  'chrII:112749-113427^113444-113450(-)')
    ('YBL040C_mRNA',   'mRNA',  'chrII:142033-142749^142846-142891(-)')
    ('YBL018C_mRNA',   'mRNA',  'chrII:185961-186352^186427-186504(-)')
    ('YBR012W-B',      'mRNA',  'chrII:259868-261173^261174-265140(+)')
    ('YBR044C_mRNA',   'mRNA',  'chrII:324292-324336^324340-326127(-)')
    ('YBR082C_mRNA',   'mRNA',  'chrII:406506-407027^407122-407379(-)')
    ('YBR126W-B_mRNA', 'mRNA',  'chrII:490824-491202(+)')
    ('YBR138C_mRNA',   'mRNA',  'chrII:513636-515391(-)')
    [rest of output omitted]


See Also
--------
`GFF3 specification <http://song.sourceforge.net/gff3.shtml>`_
    GFF3 specification by the Sequence Ontology consortium

`GTF2.2 specification <http://mblab.wustl.edu/GTF22.html>`_
    Hosted by the Brent lab

`UCSC file format FAQ <http://genome.ucsc.edu/FAQ/FAQformat.html>`_.
    GFF & GTF descriptions at UCSC
"""
__author__ = "joshua"
__date__ = "$Dec 1, 2010 11:00:55 AM$"
import itertools
import gc
import copy
import sys
from abc import abstractmethod
from plastid.util.io.filters import AbstractReader, SkipBlankReader
from plastid.util.io.openers import NullWriter, multiopen
from plastid.readers.common import get_identical_attributes, \
                                   AssembledFeatureReader
from plastid.genomics.roitools import Transcript, SegmentChain, \
                                      GenomicSegment, add_three_for_stop_codon
from plastid.readers.gff_tokens import parse_GFF3_tokens, parse_GTF2_tokens
from plastid.util.services.exceptions import DataWarning, warn



#===============================================================================
# INDEX: SO v2.5.3 feature types
#   see: http://www.sequenceontology.org/resources/intro.html
#===============================================================================

_DEFAULT_GFF3_GENE_TYPES = {
    "gene",
    "candidate_gene",
    "functional_candidate_gene",
    "positional_candidate_gene",
    "cryptic_gene",
    "cryptogene",
    "engineered_gene",
    "engineered_foreign_gene",
    "engineered_foreign_transposable_element_gene",
    "engineered_fusion_gene"
    "epigenetically_modified_gene",
    "allelically_excluded_gene",
    "gene_rearranged_at_DNA_level",
    "maternally_imprinted_gene",
    "paternally_imprinted_gene",
    "foreign_gene",
    "fusion_gene",
    "gene_cassette",
    "gene_with_non_canonical_start_codon",
    "gene_with_start_codon_CUG",
    "gene_with_polycistronic_transcript",
    "gene_with_dicistronic_transcript",
    "gene_with_dicistronic_mRNA",
    "gene_with_dicistronic_primary_transcript",
    "gene_with_trans_spliced_transcript",
    "mt_gene",
    "kinetoplast_gene",
    "maxicircle_gene",
    "minicircle_gene",
    "ncRNA_gene",
    "gRNA_gene",
    "lincRNA_gene",
    "miRNA_gene",
    "piRNA_gene",
    "RNase_MRP_RNA_gene",
    "RNase_P_RNA_gene",
    "rRNA_gene",
    "scRNA_gene",
    "snoRNA_gene",
    "snRNA_gene",
    "SRP_RNA_gene",
    "telomerase_RNA_gene",
    "tmRNA_gene",
    "tRNA_gene",
    "negatively_autoregulated_gene",
    "nuclear_gene",
    "nucleomorph_gene",
    "plasmid_gene",
    "plastid_gene",
    "apicoplast_gene",
    "chromoplast_gene",
    "ct_gene",
    "cyanelle_gene",
    "leucoplast_gene",
    "proplastid_gene",
    "positively_autoregulated_gene",
    "post_translationally_regulated_gene",
    "predicted_gene",
    "protein_coding_gene",
    "gene_with_edited_transcript",
    "gene_with_mRNA_with_frameshift",
    "gene_with_polyadenylated_mRNA",
    "gene_with_recoded_mRNA",
    "gene_with_mRNA_recoded_by_translational_bypass",
    "gene_with_stop_codon_read_through",
    "gene_with_stop_codon_redefined_as_pyrrollysine",
    "gene_with_stop_codon_redefined_as_selenocysteine",
    "gene_with_transcript_with_translational_frameshift", "proviral_gene",
    "endogenous_retroviral_gene", "non_functional_homolog_of_pseudogene",
    "non_processed_pseudogene", "cassette_pseudogene", "duplicated_pseudogene",
    "nuclear_mt_pseudogene", "pseudogene_by_unequal_crossing_over", "unitary_pseudogene",
    "polymorphic_pseudogene", "processed_pseudogene", "transposable_element_pseudogene",
    "recombinatorially_rearranged_gene", "recombinatorially_inverted_gene",
    "recombinatorially_rearranged_vertebrate_immune_system_gene", "rescue_gene",
    "wild_type_rescue_gene", "retrogene", "silenced_gene", "gene_silenced_by_DNA_modification",
    "gene_silenced_by_DNA_methylation", "gene_silenced_by_histone_modification",
    "gene_silenced_by_histone_deacetylation", "gene_silenced_by_histone_methylation",
    "gene_silenced_by_RNA_interference", "transgene", "floxed_gene",
    "translationally_regulated_gene", "transposable_element_gene",
    "engineered_foreign_transposable_element_gene",
}
"""GFF3 gene types as annotated by `SO 2.5.3 <http://www.sequenceontology.org/resources/intro.html>`_"""

_DEFAULT_GFF3_TRANSCRIPT_TYPES = {
    "transcript",
    "mature_transcript",
    "enzymatic_RNA",
    "mRNA",
    "mRNA_with_frameshift",
    "mRNA_with_minus_1_frameshift",
    "mRNA_with_minus_2_frameshift",
    "mRNA_with_plus_1_frameshift",
    "mRNA_with_plus_2_frameshift",
    "polyadenylated_mRNA",
    "polycistronic_mRNA",
    "dicistronic_mRNA",
    "monocistronic_mRNA",
    "recoded_mRNA",
    "mRNA_recoded_by_codon_redefinition",
    "mRNA_recoded_by_translational_bypass",
    "trans_spliced_mRNA",
    "ncRNA",
    "antisense_RNA",
    "MicF_RNA",
    "class_I_RNA",
    "class_II_RNA",
    "enhancerRNA",
    "guide_RNA",
    "lnc_RNA",
    "antisense_lncRNA",
    "intronic_lncRNA",
    "lincRNA",
    "piRNA",
    "priRNA",
    "rasiRNA",
    "RNase_MRP_RNA",
    "RNase_P_RNA",
    "rRNA",
    "large_subunit_rRNA",
    "rRNA_21S",
    "rRNA_23S",
    "rRNA_25S",
    "rRNA_28S",
    "rRNA_5_8S",
    "rRNA_5S",
    "small_subunit_rRNA",
    "rRNA_16S",
    "rRNA_18S",
    "rRNA_cleavage_RNA",
    "scRNA",
    "shRNA",
    "siRNA",
    "small_regulatory_ncRNA",
    "CsrB_RsmB_RNA",
    "DsrA_RNA",
    "GcvB_RNA",
    "IoR",
    "miRNA",
    "moR",
    "OxyS_RNA",
    "RNA_6S",
    "RprA_RNA",
    "RRE_RNA",
    "spot42_RNA",
    "tmRNA",
    "snoRNA",
    "C_D_box_snoRNA",
    "methylation_guide_snoRNA",
    "U14_snoRNA",
    "U3_snoRNA",
    "H_ACA_box_snoRNA",
    "pseudouridylation_guide_snoRNA",
    "snRNA",
    "U11_snRNA",
    "U12_snRNA",
    "U1_snRNA",
    "U2_snRNA",
    "U4_snRNA",
    "U4atac_snRNA",
    "U5_snRNA",
    "U6_snRNA",
    "U6atac_snRNA",
    "SRP_RNA",
    "tasiRNA",
    "telomerase_RNA",
    "telomeric_transcript",
    "anti_ARRET",
    "ARIA",
    "ARRET",
    "TERRA",
    "tRNA",
    "alanyl_tRNA",
    "arginyl_tRNA",
    "asparaginyl_tRNA",
    "aspartyl_tRNA",
    "cysteinyl_tRNA",
    "glutaminyl_tRNA",
    "glutamyl_tRNA",
    "glycyl_tRNA",
    "histidyl_tRNA",
    "isoleucyl_tRNA",
    "leucyl_tRNA",
    "lysyl_tRNA",
    "methionyl_tRNA",
    "phenylalanyl_tRNA",
    "prolyl_tRNA",
    "pyrrollysyl_tRNA",
    "selenocysteinyl_tRNA",
    "seryl_tRNA",
    "threonyl_tRNA",
    "tryptophanyl_tRNA",
    "tyrosyl_tRNA",
    "valyl_tRNA",
    "vault_RNA",
    "Y_RNA",
    "edited_transcript",
    "edited_mRNA",
    "edited_transcript_by_A_to_I_substitution",
    "non_functional_homolog_of_pseudogenic_transcript",
    "pseudogenic_transcript",
}
"""GFF3 mature transcript types as annotated by `SO 2.5.3
<http://www.sequenceontology.org/resources/intro.html>`_"""

_DEFAULT_GFF3_EXON_TYPES = {
    "exon",
    "coding_exon",
    "noncoding_exon",
    "exon_of_single_exon_gene",
    "interior_exon",
    "interior_coding_exon",
    "five_prime_coding_exon",
    "three_prime_coding_exon"
    "five_prime_noncoding_exon",
    "three_prime_noncoding_exon",
    "pseudogenic_exon",
}
"""GFF3 exon feature types as annotated by `SO 2.5.3
<http://www.sequenceontology.org/resources/intro.html>`_"""

_DEFAULT_GFF3_CDS_TYPES = {
    "CDS",
    "CDS_fragment",
    "CDS_indpendently_known",
    "CDS_predicted",
}
"""GFF3 CDS feature types as annotated by `SO 2.5.3
<http://www.sequenceontology.org/resources/intro.html>`_"""

#===============================================================================
# INDEX: Readers for GFF formats
#===============================================================================

StopFeature = SegmentChain(GenomicSegment("Stop", 0, 1, "."), type="StopFeature", ID="StopFeature")
"""Special |SegmentChain| emitted from GFF readers when:

- the special line ``###`` is encountered
- the special line ``###FASTA`` is encountered
- a GFF file is marked as sorted, and the contig/chromosome changes
- the source stream of features is changed

indicating that all previously returned features may be assembled into full
objects.

.. note::

   Because :obj:`StopFeature` is zero-length, it does not evaluate as equal to
   itself. Use ``x is StopFeature`` or ``x is not StopFeature`` it testing for
   equality.
"""


class AbstractGFF_Reader(AbstractReader):
    """Abstract base class for GFF readers.

    Parses GFF streams line by line into |GenomicSegment|

    Attributes
    ----------
    metadata : dict
        Dictionary of metadata found in file headers
    """

    def __init__(self, *streams, **kwargs):
        """Create an |AbstractGFF_Reader|

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to GFF information

        adjust_to_0 : bool, optional
            Boolean, whether or not to adjust feature
            indices to a 0 base. True for `GTF2`_ and `GFF3`_
            files, as these are 1-indexed. (Default: `True`)

        end_included : bool, optional
            Boolean, whether the end coordinate is
            included in the feature (closed or 'end-included' intervals)
            or not (half-open intervals). (Default: `True`)

        return_stopfeatures : bool, optional
            If `True`, will return a special |SegmentChain| called
            :obj:`StopFeature` signifying that all previously emitted GFF
            entries may be assembled into complete entities. These are emitted
            when the line `'###'` is encountered in a GFF. (Default: `True`)

        is_sorted : bool, optional
            If True and `return_stopfeatures` is True, assume the GFF is sorted.
            The reader will return :obj:`StopFeature` when the chromosome name
            of a given feature differs from that of the previous feature.
            (Default: `False`)

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)
        """
        stream = itertools.chain.from_iterable(multiopen(streams, fn=open))
        if kwargs.get("tabix", False) == True:
            stream = ("\t".join(X) for X in stream)

        self.chromosomes = {}
        self.metadata = {}
        self.line_queue = []

        self.adjust_to_0 = kwargs.get("adjust_to_0", True)
        self.end_included = kwargs.get("end_included", True)
        self.return_stopfeatures = kwargs.get("return_stopfeatures", True)
        self.is_sorted = kwargs.get("is_sorted", False)

        line = next(stream)
        while line[0:2] == "##":
            self._parse_metatokens(line[2:])
            line = next(stream)

        self.line_queue.append(line)
        self._last_chrom = None

        self.stream = itertools.chain(self.line_queue, SkipBlankReader(stream))
        super(AbstractGFF_Reader, self).__init__(self.stream)

    def _parse_metatokens(self, inp):
        """Parses metadata embedded in a GFF stream, and stores
        these in appropriate attributes.

        Parameters
        ----------
        inp : str
            line of GFF input

        """
        items = inp.rstrip().split()
        if len(items) > 0:
            key = items[0]
            if key == "sequence-region":
                try:
                    self.chromosomes[items[1]] = (items[2], items[3])
                except IndexError:
                    self.chromosomes[items[1]] = tuple(items[1:])

            elif key in self.metadata.keys():
                self.metadata[key] += ";" + " ".join(items[1:])

            else:
                self.metadata[key] = " ".join(items[1:])

    @abstractmethod
    def _parse_tokens(self, attr_string):
        """Placeholder function to parse column 9, which is formatted
        differently in different GFF subtypes. Implement this
        in subclasses

        Parameters
        ----------
        attr_string : str
            Ninth column of GFF

        Returns
        -------
        dict
            Dictionary of parsed tokens from ninth GFF column
        """
        pass

    def _parse_genomic_feature(self, line):
        """Parse GFF lines into |SegmentChain| objects

        Parameters
        ----------
        line : str
            Valid line of a GFF formatted file

        Returns
        -------
        |SegmentChain|
        """
        # yapf: disable
        items = line.rstrip("\n").split("\t")
        chrom        = items[0]
        source       = items[1]
        feature_type = items[2]
        start        = int(items[3]) - int(self.adjust_to_0)
        end          = int(items[4]) - int(self.adjust_to_0) + int(self.end_included)
        score        = items[5]
        strand       = items[6]
        phase        = items[7]
        attr_string  = items[8]

        info_dict = self._parse_tokens(attr_string)
        info_dict["source"] = source
        info_dict["score"]  = score
        info_dict["phase"]  = phase
        info_dict["type"]   = feature_type
        # yapf: enable

        my_iv = GenomicSegment(chrom, start, end, strand)
        my_feature = SegmentChain(my_iv, **info_dict)

        if chrom != self._last_chrom:
            old_chrom = self._last_chrom
            self._last_chrom = chrom
            if old_chrom is not None:
                if self.is_sorted == True and self.return_stopfeatures == True:
                    self.stream = itertools.chain([line], self.stream)
                    return StopFeature

        return my_feature

    def filter(self, line):
        """Parses lines of the GFF stream into |SegmentChain|
        When metadata is found, temporarily delegates processing to
        :meth:`_parse_metatokens`, and then reads the next genomic feature

        Parameters
        ----------
        line
            Next line from GFF stream

        Returns
        -------
        |SegmentChain|
            Next feature in file
        """
        if line.startswith("###"):
            if self.return_stopfeatures == True:
                return StopFeature
            else:
                return next(self)

        elif line.startswith("##FASTA"):
            return StopFeature

        elif line.startswith("##"):
            self._parse_metatokens(line[2:])
            return next(self)

        elif line.startswith("#"):
            return next(self)

        else:
            return self._parse_genomic_feature(line)


class GFF3_Reader(AbstractGFF_Reader):
    """
    GFF3_Reader(*streams, end_included=True, return_stopfeatures=False, is_sorted=False, tabix=False)

    Read raw features in `GFF3`_ files as |SegmentChains|.

    Users who wish to reconstruct |Transcripts| from raw features should instead
    use |GFF3_TranscriptAssembler|, which performs this task automatically.

    Assumes input stream to use 1-indexed coordinates, in compliance with the
    `Sequence Ontology GFF3 specification <http://song.sourceforge.net/gff3.shtml>`_.

    `GFF3`_ attributes (from column 9) for each record are stored in its ``attr``
    dictionary. Names and values of attributes are unescaped. The values for the
    attributes `Parent`, `Alias`, `Dbxref`, `dbxref`, and `Note`, if present, are
    lists rather than strings, because the `GFF3`_ spec enables these to have
    multiple values.


    Parameters
    ----------
    *streams : one or more str or file-like
        One or more input streams or filenames pointing to GFF information

    end_included : bool, optional
        Boolean, whether the end coordinate is included in the feature (closed
        or 'end-included' intervals) or not (half-open intervals). All
        coordinates will be normalized to 0-indexed, half-open (Default: `True`)

    return_stopfeatures : bool, optional
        If `True`, return a special |SegmentChain| called :py:obj:`StopFeature`
        signifying that all previously emitted GFF entries may be assembled
        into complete entities. These are emitted when the line "###"
        is encountered in a `GFF3`_. (Default: `False`)

    is_sorted : bool, optional
        If `True` and `return_stopfeatures` is `True`, assume the `GFF3`_ is
        sorted.  The reader will return :obj:`StopFeature` when the chromosome
        name of a given feature differs from that of the previous feature.
        (Default: `False`)

    tabix : boolean, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)


    Attributes
    ----------
    metadata : dict
        Dictionary of metadata found in file headers


    Examples
    --------

    Read raw features from a `GFF3`_ file::

        >>> feature_reader = GFF3_Reader(open("./some_file.gff"))
        >>> for feature in feature_reader:
        >>>     print(feature.get_name(), feature.attr["type"], str(feature))
        ('chrI', 'chromosome', 'chrI:0-230218(.)')
        ('TEL01L-TR', 'telomeric_repeat', 'chrI:0-62(-)')
        ('TEL01L', 'telomere', 'chrI:0-801(-)')
        ('TEL01L-XR', 'X_element_combinatorial_repeat', 'chrI:62-336(-)')
        ('YAL069W', 'gene', 'chrI:334-649(+)')
        ('TEL01L-XC', 'X_element', 'chrI:336-801(-)')
        ('TEL01L-XC_nucleotide_match', 'nucleotide_match', 'chrI:752-763(-)')
        ('TEL01L-XC_binding_site', 'binding_site', 'chrI:531-544(-)')
        ('YAL068W-A', 'gene', 'chrI:537-792(+)')
        ('ARS102', 'ARS', 'chrI:649-1791(.)')
        [rest of output omitted]
    """

    def __init__(self, *streams, **kwargs):
        """
        GFF3_Reader(*streams, end_included=True, return_stopfeatures=False, is_sorted=False, tabix=False)

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to GFF information

        end_included : bool, optional
            Boolean, whether the end coordinate is
            included in the feature (closed or 'end-included' intervals)
            or not (half-open intervals). (Default: `True`)

        return_stopfeatures : bool, optional
            If `True`, return a special |SegmentChain| called :py:obj:`StopFeature`
            signifying that all previously emitted GFF entries may be assembled
            into complete entities. These are emitted when the line "###"
            is encountered in a `GFF3`_. (Default: `False`)

        is_sorted : bool, optional
            If `True` and `return_stopfeatures` is `True`, assume the `GFF3`_ is
            sorted.  The reader will return :obj:`StopFeature` when the
            chromosome name of a given feature differs from that of the previous
            feature.  (Default: `False`)

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)
         """
        super(GFF3_Reader, self).__init__(*streams, adjust_to_0=True, **kwargs)

    def _parse_tokens(self, inp):
        """Parse column 9 of `GFF3`_ into attribute dictionary

        Parameters
        ----------
        inp : str
            Ninth column of `GFF3`_

        Returns
        -------
        dict
            Dictionary of parsed tokens from ninth `GFF3`_ column
        """
        return parse_GFF3_tokens(inp)


class GTF2_Reader(AbstractGFF_Reader):
    """
    GTF2_Reader(*streams, end_included=True, return_stopfeatures=False, is_sorted=False, tabix=False)

    Read raw features in `GTF2`_ files as |SegmentChains|. To assemble transcripts
    from raw features, use |GTF2_TranscriptAssembler|.

    Assumes input to comply with the
    `GTF2 specification <http://mblab.wustl.edu/GTF22.html>`_. Each element must:

      - use 1-indexed, fully-closed coordinates
      - have defined `gene_id` and `transcript_id` attributes

    All |SegmentChain| objects returned by the reader have 0-indexed,
    half-open coordinates in keeping with Python conventions.


    Parameters
    ----------
    *streams : one or more str or file-like
        One or more input streams or filenames pointing to GFF information

    end_included : bool, optional
        Boolean, whether the end coordinate is included in the feature (closed
        or 'end-included' intervals) or not (half-open intervals).
        (Default: `True`)

    return_stopfeatures : bool, optional
        If `True`, will return a special |SegmentChain| called
        :py:obj:`StopFeature` signifying that all previously emitted
        SegmentChains may be assembled into complete entities. These are emitted
        when the line "###" is encountered in a `GTF2`_. (Default: `False`)

    is_sorted : bool, optional
        If `True` and `return_stopfeatures` is `True`, assume the `GTF2`_ is
        sorted by chromosome. The reader will return :obj:`StopFeature` when the
        chromosome name of a given feature differs from that of the previous
        feature. (Default: `False`)

    tabix : boolean, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)

    Examples
    --------
    Read raw features from a `GTF2`_ file::

        >>> feature_reader = GTF2_Reader(open("some_file.gtf"))
        >>> for feature in reader:
        >>>     print(feature.get_name(),feature.attr["type"],str(feature))
        ('YAL030W_mRNA',  'exon',        'chrI:87262-87387(+)')
        ('YAL030W_mRNA',  'exon',        'chrI:87500-87857(+)')
        ('YAL030W_mRNA',  'CDS',         'chrI:87285-87387(+)')
        ('YAL030W_mRNA',  'CDS',         'chrI:87500-87749(+)')
        ('YAL030W_mRNA',  'start_codon', 'chrI:87285-87288(+)')
        ('YAL030W_mRNA',  'stop_codon',  'chrI:87749-87752(+)')
        ('YBL092W_mRNA',  'exon',        'chrII:45643-45644(+)')
        ('YBL092W_mRNA',  'exon',        'chrII:45977-46440(+)')
        ('YBL092W_mRNA',  'CDS',         'chrII:45977-46367(+)')
        ('YBL092W_mRNA',  'start_codon', 'chrII:45977-45980(+)')
        [rest of output omitted]


    Attributes
    ----------
    metadata : dict
        Dictionary of metadata found in file headers
    """

    def __init__(self, *streams, **kwargs):
        """
        GTF2_Reader(*streams, end_included=True, return_stopfeatures=False, is_sorted=False, tabix=False)

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to GFF information

        end_included : bool, optional
            Boolean, whether the end coordinate is
            included in the feature (closed or 'end-included' intervals)
            or not (half-open intervals). (Default: `True`)

        return_stopfeatures : bool, optional
            If `True`, will return a special |SegmentChain| called
            :py:obj:`StopFeature` signifying that all previously emitted
            SegmentChains may be assembled into complete entities. These are
            emitted when the line "###" is encountered in a `GTF2`_. (Default:
            `False`)

        is_sorted : bool, optional
            If `True` and `return_stopfeatures` is `True`, assume the `GTF2`_ is
            sorted by chromosome. The reader will return :obj:`StopFeature` when
            the chromosome name of a given feature differs from that of the
            previous feature. (Default: `False`)

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)
        """
        super(GTF2_Reader, self).__init__(*streams, adjust_to_0=True, **kwargs)

    def _parse_tokens(self, inp):
        """Parse column 9 of `GTF2`_ into dictionary

        Parameters
        ----------
        inp : str
            Ninth column of `GTF2`_

        Returns
        -------
        dict
            Dictionary of parsed tokens from ninth `GTF2`_ column
        """
        return parse_GTF2_tokens(inp)


class AbstractGFF_Assembler(AssembledFeatureReader):
    """Abstract base class for readers that assemble composite features
    -- e.g. |Transcript| objects -- from one or more features in one or
    more streams of `GTF2`_ or `GFF3`_ data.


    Attributes
    ----------
    stream : file-like
        Input stream, usually constructed from or more open filehandles

    metadata : dict
        Various attributes gleaned from the stream, if any

    counter : int
        Cumulative line number counter over all streams

    printer : file-like, optional
        Logger implementing a ``write()`` method.

    rejected : list
        A list of transcript IDs that failed to assemble properly
    """

    def __init__(self, *streams, **kwargs):
        """Create a |AbstractGFF_Assembler|

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to GFF information

        is_sorted : bool, optional
            GFF is sorted by chromosome name, allowing some memory savings
            (Default: `False`)

        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default:
            |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations.
            If set to `True`, three nucleotides will be added to the threeprime
            end of each CDS annotation. (Default: `False`)

        printer : file-like, optional
            Logger implementing a ``write()`` method. (Default: |NullWriter|)

        reader_class : class
            |GFF3_Reader| or |GTF2_Reader|

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)

        **kwargs
            Other keyword arguments used by specific parsers
        """

        tabix = kwargs.get("tabix", False)
        end_included = kwargs.get("end_included", True)
        return_stopfeatures = kwargs.get("return_stopfeatures", True)
        is_sorted = kwargs.get("is_sorted", False)
        reader_class = kwargs.get("reader_class")

        streams = multiopen(streams, open)

        iterables = []
        for stream in streams:
            iterables.append(
                reader_class(
                    stream,
                    end_included=end_included,
                    return_stopfeatures=return_stopfeatures,
                    is_sorted=is_sorted,
                    tabix=tabix
                )
            )
            iterables.append([StopFeature])

        self.stream = itertools.chain.from_iterable(iterables)

        self.printer = kwargs.get("printer", NullWriter())
        self.return_type = kwargs.get("return_type", SegmentChain)
        self.add_three_for_stop = kwargs.get("add_three_for_stop", False)

        self.metadata = {}
        self.rejected = []
        self.counter = 0

        self._transcript_cache = iter([])
        self._feature_cache = {}

    def _finalize(self, tx):
        return tx

    @abstractmethod
    def _collect(self, feature):
        """Collect relevant features of transcripts

        Parameters
        ----------
        feature : |SegmentChain|
            Feature to collect
        """
        pass

    @abstractmethod
    def _assemble_transcripts(self):
        """Assemble |Transcript| objects from collected features

        Returns
        -------
        list
            list of transcripts
        """
        pass

    @abstractmethod
    def _reset(self):
        """Release memory and reset internal hashes"""
        pass

    def _get_transcript_batches(self):
        """Lazily assemble batches of transcripts from components in
        `self.stream`, when signals in `self.stream` indicate it is safe to
        assemble transcripts from previously collected features. This feature is
        called by `self.__iter__`, which returns individual transcripts.

        Signals that can signify safety to assemble previous transcripts
        include:

         - The special line ``###``, which indicates that the preceding
           collection of lines in a GFF3 file are safe to assemble

         - The special line ``###FASTA`` which indicates end of the feature
           section of a GFF3 file

         - A change in reference files (if multiple input files were given)

        When an assembly signal is given, transcripts are assembled an internal
        caches of collected feature are reset to free memory.


        Yields
        ------
        list
            Sorted list of transcripts in current GFF block
        """
        for feature in self.stream:
            self.counter += 1

            # collect features until a stop signal, such as:
            #     - change of contig if GTF2/GFF3 is sorted
            #     - '###FASTA' block, indicating end of features
            #     - switch between source files in `self.stream`
            if feature is not StopFeature:
                self._collect(feature)

            # if stop signal is reached, clear memory, because those features
            # no longer needed
            else:
                self.printer.write("Assembling next batch of transcripts ...")
                transcripts, rejected = self._assemble_transcripts()

                # remove previously assembled transcripts
                # these lines required to free memory in Python 2.7
                # and early versions of Python 3.x
                gc.collect()
                del gc.garbage[:]

                # sort and prep next set of transcripts
                transcripts = sorted(transcripts)
                self.rejected.extend(rejected)
                self._reset()
                yield transcripts

        # assemble any remaining features into transcripts, e.g. if no
        # stop signals present within a single file
        transcripts, rejected = self._assemble_transcripts()
        gc.collect()
        del gc.garbage[:]

        transcripts = sorted(transcripts)
        self.rejected.extend(rejected)
        self._reset()
        yield transcripts

    def __iter__(self):
        """Return next assembled transcript from GFF

        Yields
        ------
        |Transcript|
            Next complex feature in annotation (usually a transcript)
        """
        for my_batch in self._get_transcript_batches():
            for my_tx in my_batch:
                yield self._finalize(my_tx)


class GTF2_TranscriptAssembler(AbstractGFF_Assembler):
    """
    GTF2_TranscriptAssembler(*streams, is_sorted=False, return_type=SegmentChain, add_three_for_stop=False, printer=None, tabix=False)

    Assemble |Transcripts| from raw features in `GTF2`_ format.

    Exons and CDS features are grouped by shared ``transcript_id``.
    Attributes that have common values for all exons and CDS within a transcript
    are propagated up to the `attr` dict of the assembled |Transcript|. Other
    attributes from individual CDS or exon components are discarded.

    The assembler functions as an iterator. Within each chromosome, transcripts
    are returned in lexical order.

    For access to raw features, instead use |GTF2_Reader|.


    Parameters
    ----------
    *streams : one or more str or file-like
        One or more input streams or filenames pointing to `GTF2`_ data

    is_sorted : bool, optional
        `GTF2`_ is sorted by chromosome name, allowing some memory savings
        (Default: `False`)

    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default:
        |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If
        set to `True`, three nucleotides will be added to the threeprime end of
        each CDS annotation, UNLESS the annotated transcript contains explicit
        `stop_codon` feature. (Default: `False`)

    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|

    tabix : boolean, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)


    Examples
    --------
    Assemble transcripts from a `GTF2`_ file::

        >>> transcript_reader = GTF2_TranscriptAssembler(open("some_file.gtf"))
        >>> for transcript in reader:
        >>>     print(transcript.get_name(),transcript.attr["type"],str(transcript)) # do something

        ('YAL030W_mRNA',   'mRNA',  'chrI:87262-87387^87500-87857(+)')
        ('YBL092W_mRNA',   'mRNA',  'chrII:45643-45644^45977-46440(+)')
        ('YBL057C_mRNA',   'mRNA',  'chrII:112749-113427^113444-113450(-)')
        ('YBL040C_mRNA',   'mRNA',  'chrII:142033-142749^142846-142891(-)')
        ('YBL018C_mRNA',   'mRNA',  'chrII:185961-186352^186427-186504(-)')
        ('YBR012W-B',      'mRNA',  'chrII:259868-261173^261174-265140(+)')
        ('YBR044C_mRNA',   'mRNA',  'chrII:324292-324336^324340-326127(-)')
        ('YBR082C_mRNA',   'mRNA',  'chrII:406506-407027^407122-407379(-)')
        ('YBR126W-B_mRNA', 'mRNA',  'chrII:490824-491202(+)')
        ('YBR138C_mRNA',   'mRNA',  'chrII:513636-515391(-)')
        [rest of output omitted]


    Attributes
    ----------
    streams : file-like
        Input streams, usually constructed from one or more open filehandles

    metadata : dict
        Various attributes gleaned from the streams, if any

    counter : int
        Cumulative line number counter over all streams

    printer : file-like, optional
        Logger implementing a ``write()`` method.

    rejected : list
        A list of transcript IDs from transcripts that failed to assemble
        properly
    """
    # transcripts can be represented as collections of exons + cds
    # or cds + UTRs, et c. We consider all UTR and exons as exons
    # and CDS, and start & stop codons as CDS areas
    _feature_map = {
        "exon": ["exon_like"],
        "5UTR": ["exon_like"],
        "3UTR": ["exon_like"],
        "CDS" : ["CDS_like", "exon_like"],
        "start_codon": ["CDS_like", "exon_like"],
        "stop_codon" : ["CDS_like", "exon_like"],
    } # yapf: disable
    dtmp = {"exon_like": {}, "CDS_like": {}}

    def __init__(self, *streams, **kwargs):
        """
        GTF2_TranscriptAssembler(*streams, is_sorted=False, return_type=SegmentChain, add_three_for_stop=False, printer=None, tabix=False)

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to `GTF2`_ data

        is_sorted : bool, optional
            `GTF2`_ is sorted by chromosome name, allowing some memory savings
            (Default: `False`)

        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations. If set to
            `True`, three nucleotides will be added to the threeprime end of each
            CDS annotation, UNLESS the annotated transcript contains explicit `stop_codon`
            feature. (Default: `False`)

        printer : file-like, optional
            Logger implementing a ``write()`` method. Default: |NullWriter|

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)
        """
        AbstractGFF_Assembler.__init__(self, *streams, reader_class=GTF2_Reader, **kwargs)
        self._feature_cache = {"exon_like": {}, "CDS_like": {}}

    def _collect(self, feature):
        """Collect transcript component CDS and exons objects from
        `self.streams`, and populate `self._feature_cache`

        Parameters
        ----------
        feature : |SegmentChain|
            Feature to collect
        """
        feature_classes = self._feature_map.get(feature.attr["type"], None)
        if feature_classes is not None:
            tname = feature.attr.get("transcript_id")
            for feature_class in feature_classes:
                try:
                    self._feature_cache[feature_class][tname].append(feature)
                except KeyError:
                    self._feature_cache[feature_class][tname] = [feature]
                    sys.exc_traceback = None

    def _assemble_transcripts(self):
        """Assemble |Transcript| objects from `self._feature_cache`,
        mapping transcript IDs to corresponding CDS and exon features.

        Attributes common to all CDS and exons for a given transcript (e.g.
        `gene_id` and `transcript_id`) are propagated up to the |Transcript|.
        Other component attributes are discarded.
        """
        rejected_transcripts = []
        transcripts = []
        for tname in set(self._feature_cache["exon_like"].keys()) \
                | set(self._feature_cache["CDS_like"].keys()):
            exons = self._feature_cache["exon_like"].pop(tname, [])
            cds = self._feature_cache["CDS_like"].pop(tname, [])
            if len(exons) > 0:
                exons = sorted(exons)
                exon_segments = [X.spanning_segment for X in exons]
            elif len(cds) > 0:
                # if cds but no exons, create exons since they are implied
                exons = sorted(cds)
                exon_segments = [X.spanning_segment for X in exons]

            # propagate attributes that are the same in all exons/cds
            # to parent. This should include `gene_id` and `transcript_id`
            attr = get_identical_attributes(exons + cds)
            if len(cds) > 0:
                cds = sorted(cds)
                attr["cds_genome_end"] = cds[-1].spanning_segment.end
                attr["cds_genome_start"] = cds[0].spanning_segment.start

            try:
                my_tx = Transcript(*tuple(exon_segments), **attr)
                if self.add_three_for_stop == True:
                    # only move stop codons if no exon feature is of type "stop_codon"
                    if "stop_codon" not in set([X.attr["type"] for X in exons]):
                        my_tx = add_three_for_stop_codon(my_tx)

                transcripts.append(my_tx)
            except ValueError:
                warn(
                    "Rejecting transcript '%s' because it contains exons on "
                    "multiple chromosomes or strands." % tname, DataWarning
                )
                # transcripts with exons on two strands
                rejected_transcripts.append(tname)
            except KeyError:
                # transcripts where CDS ends outside bounds of transcript
                # there are 25 of these in flybase r5.43
                rejected_transcripts.append(tname)
                warn(
                    "Rejecting transcript '%s' because start or stop codons "
                    "are outside exon boundaries." % tname, DataWarning
                )
        transcripts.sort()
        sys.exc_traceback = None

        return transcripts, rejected_transcripts

    def _reset(self):
        """Release memory and reset internal hashes"""
        del self._feature_cache
        gc.collect()
        del gc.garbage[:]
        self._feature_cache = {"exon_like": {}, "CDS_like": {}}


class GFF3_TranscriptAssembler(AbstractGFF_Assembler):
    """
    GFF3_TranscriptAssembler(*streams, is_sorted=False, return_type=SegmentChain, add_three_for_stop=False, printer=None, tabix=False)

    Assemble |Transcripts| from raw features in `GFF3`_ format.

    Within a chromosome, transcripts are returned in lexical order. Features
    that do not constitute portions of transcripts (e.g.  origins of replication)
    are ignored. For access to those, read raw features using |GFF3_Reader|.


    Parameters
    ----------
    streams : one or more str or file-like
        One or more input streams or filenames pointing to `GFF3`_ data

    is_sorted : bool, optional
        `GFF3`_ is sorted by chromosome name, allowing some memory savings
        (Default: `False`)

    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default:
        |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If
        set to `True`, three nucleotides will be added to the threeprime end of
        each CDS annotation. (Default: `False`)

    transcript_types : list, optional
        List of `GFF3`_ feature types that should be considered as transcripts
        (Default: as specified in SO 2.5.3 )

    exon_types : list, optional
        List of `GFF3`_ feature types that should be considered as exons or
        contributing to transcript nucleotide positions
        during transcript assembly (Default: as specified in SO 2.5.3 )

    cds_types : list, optional
        List of `GFF3`_ feature types that should be considered as CDS or
        contributing to transcript coding regions during transcript assembly
        (Default: as specified in SO 2.5.3 )

    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|

    tabix : boolean, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)


    Examples
    --------
    Assemble transcripts from a `GFF3`_ file::

        >>> transcript_reader = GFF3_TranscriptAssembler(open("some_file.gff"))
        >>> for transcript in reader:
        >>>     print(transcript.get_name(),transcript.attr["type"],str(transcript)) # do something

        ('YAL030W_mRNA',   'mRNA',  'chrI:87262-87387^87500-87857(+)')
        ('YBL092W_mRNA',   'mRNA',  'chrII:45643-45644^45977-46440(+)')
        ('YBL057C_mRNA',   'mRNA',  'chrII:112749-113427^113444-113450(-)')
        ('YBL040C_mRNA',   'mRNA',  'chrII:142033-142749^142846-142891(-)')
        ('YBL018C_mRNA',   'mRNA',  'chrII:185961-186352^186427-186504(-)')
        ('YBR012W-B',      'mRNA',  'chrII:259868-261173^261174-265140(+)')
        ('YBR044C_mRNA',   'mRNA',  'chrII:324292-324336^324340-326127(-)')
        ('YBR082C_mRNA',   'mRNA',  'chrII:406506-407027^407122-407379(-)')
        ('YBR126W-B_mRNA', 'mRNA',  'chrII:490824-491202(+)')
        ('YBR138C_mRNA',   'mRNA',  'chrII:513636-515391(-)')
        [rest of output omitted]


    Attributes
    ----------
    streams : file-like
        Input stream, usually constructed from or more open filehandles

    metadata : dict
        Various attributes gleaned from the stream, if any

    counter : int
        Cumulative line number counter over all streams

    printer : file-like, optional
        Logger implementing a ``write()`` method.

    rejected : list
        A list of transcript IDs from transcripts that failed to assemble
        properly


    Notes
    -----
   `GFF3`_ schemas vary
       `GFF3`_ files can have many different schemas of hierarchy. We deal with
       that here by allowing users to supply `transcript_types` and
       `exon_types`, to indicate which sorts of features should be included. By
       default, we use a subset of the schema set out in `Seqence Ontology 2.5.3
       <http://www.sequenceontology.org/resources/intro.html>`_

       Briefly:

        1. The GFF3 file is combed for transcripts of the types specified by
        `transcript_types`, exons specified by `exon_types`, and CDS specified
        by types listed in `cds_types`.

        2. Exons and CDS are matched with their parent transcripts by matching
           the `Parent` attributes of CDS and exons to the `ID` of transcripts.
           Transcripts are then constructed from those intervals, and coding
           regions set accordingly.

        3. If exons and/or CDS features point to a `Parent` that is not
           in `transcript_types`, they are grouped into a new transcript,
           whose ID is set to the value of their shared `Parent`. However,
           this value for `Parent` might refer to a gene rather than
           a transcript; unfortunately this cannot be known without other
           information. Attributes that are common to all CDS and exon
           features are bubbled up to the transcript.

        4. If exons and/or CDS features have no `Parent`, but share a common ID,
           they are grouped by ID into a single transcript. Attributes common
           to all CDS and exon features are bubbled up to the transcript.
           The `Parent` attribute is left unset.

        5. If a transcript feature is annotated but has no child CDS or exons,
           the transcript is assumed to be non-coding and is assembled from
           any transcript-type features that share its `ID` attribute.

   Identity relationships between elements vary between `GFF3`_ files
       Different `GFF3`_ files specify discontiguous features differently. For
       example, in `Flybase`_, different exons of a transcript will have unique
       IDs, but will share the same `'Parent'` attribute in column 9 of the GFF.
       In `Wormbase`_, however, different exons of the same transcript will
       share the same ID. Here, we first check for the Flybase style (by
       Parent), then fall back to Wormbase style (by shared ID).

   Transcript assembly
       To save memory, transcripts are assembled lazily as follows:

       #.  If there exist assembled transcripts in `self._transript_cache`,
           return the next transcript. Transcripts in the cache are stored
           lexically.

       #.  Otherwise, collect features from the `GFF3`_ stream until either a
           `'###'` line or EOF is encountered. Then, assemble transcripts and
           store them in `self._transcript_cache`. Delete unused features
           from memory. If the `GFF3`_ is sorted, then a change in chromosome
           name will also trigger assembly of collected features.
    """

    def __init__(self, *streams, **kwargs):
        """
        GFF3_TranscriptAssembler(*streams, is_sorted=False, return_type=SegmentChain, add_three_for_stop=False, printer=None, tabix=False)

        Parameters
        ----------
        *streams : one or more str or file-like
            One or more input streams or filenames pointing to `GFF3`_ data

        is_sorted : bool, optional
            `GFF3`_ is sorted by chromosome name, allowing some memory savings
            (Default: `False`)

        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default:
            |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations.
            If set to `True`, three nucleotides will be added to the threeprime
            end of each CDS annotation. (Default: `False`)

        transcript_types : list, optional
            List of `GFF3`_ feature types that should be considered as
            transcripts (Default: as specified in SO 2.5.3 )

        exon_types : list, optional
            List of `GFF3`_ feature types that should be considered as exons or
            contributing to transcript nucleotide positions
            during transcript assembly (Default: as specified in SO 2.5.3 )

        cds_types : list, optional
            List of `GFF3`_ feature types that should be considered as CDS or
            contributing to transcript coding regions during transcript assembly
            (Default: as specified in SO 2.5.3 )

        printer : file-like, optional
            Logger implementing a ``write()`` method. Default: |NullWriter|

        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)


        Notes
        -----
        Sequence Ontology 2.5.3
            By default, this assembler constructs transcripts following a subset
            of the `GFF3`_ schema from the `SO Consortium
            <http://www.sequenceontology.org/resources/intro.html>`_.  For
            details on assembly see the :class:`class docstring
            <GFF3_TranscriptAssembler>`, above.
        """
        AbstractGFF_Assembler.__init__(self, *streams, reader_class=GFF3_Reader, **kwargs)
        self.transcript_types = set(kwargs.get("transcript_types", _DEFAULT_GFF3_TRANSCRIPT_TYPES))
        self.exon_types = set(kwargs.get("exon_types", _DEFAULT_GFF3_EXON_TYPES))
        self.cds_types  = set(kwargs.get("cds_types", _DEFAULT_GFF3_CDS_TYPES)) # yapf: disable
        self.transcript_components = self.exon_types | self.cds_types
        self._feature_cache = {}
        self._tx_features = {}
        self._reset()

    def _collect(self, feature):
        """Collect CDS and exon components of transcripts from `self.streams`,
        and populate `self._feature_cache`

        Parameters
        ----------
        feature : |SegmentChain|
            Feature to collect
        """
        feature_name = feature.get_name()

        if feature.attr["type"] in self.transcript_types:
            try:
                self._tx_features[feature_name].append(feature)
            except KeyError:
                self._tx_features[feature_name] = [feature]
                sys.exc_traceback = None

        elif feature.attr["type"] in self.transcript_components:
            # assume parent is transcript ID.
            # If no Parent, assume transcript is described as exon or CDS
            # with identical ID attributes

            tnames = feature.attr.get("Parent", [feature.attr.get("ID", [])])

            if len(tnames) == 0:
                warn(
                    "Found %s at %s with no `Parent` or `ID`. Ignoring." %
                    (feature.attr["type"], str(feature.spanning_segment)), DataWarning
                )
            for tname in tnames:
                try:
                    self._feature_cache[feature.attr["type"]][tname].append(feature)
                except KeyError:
                    self._feature_cache[feature.attr["type"]][tname] = [feature]
                    sys.exc_traceback = None

    def _assemble_transcripts(self):
        """Assemble |Transcript| s from components in `self._feature_cache`

        Returns
        -------
        list
            list of transcripts
        """
        rejected = []
        transcripts = []
        tx_features_counted = []

        # names of transcripts in transcript_types
        tnames = set([])

        # find parents of exons & cds that are not present
        # in types from transcript_types
        for type_ in self.exon_types | self.cds_types:
            tnames |= set(self._feature_cache[type_].keys())

        for tname in tnames:
            tx_features_counted.append(tname)

            exons = []
            for type_ in self.exon_types:
                exons.extend(self._feature_cache[type_].get(tname, []))

            cds = []
            for cds_type in self.cds_types:
                cds.extend(self._feature_cache[cds_type].get(tname, []))

            # if transcript is represented, not just implied, use its attributes
            if tname in self._tx_features:

                # use transcript name as gene if no Parent
                gene_id = self._tx_features[tname][0].attr.get("Parent", [tname])

                # gene IDs are now returned as lists from GFF3 parser
                gene_id = ",".join(sorted(gene_id))

                # get attr from transcript object
                attr = self._tx_features[tname][0].attr
                attr["ID"] = tname
                attr["transcript_id"] = tname
                attr["gene_id"] = gene_id

            # if transcript is just implied by presence of CDS and exons
            # TODO: make sure "Parent" will carry over sensibly from lists
            else:
                attr = get_identical_attributes(exons + cds)
                attr["ID"] = tname
                attr["transcript_id"] = tname
                attr["type"] = "mRNA"

            exon_segments = [X.spanning_segment for X in exons]
            cds_segments = [X.spanning_segment for X in cds]

            if len(exon_segments) + len(cds_segments) > 0:
                # transcript with child features
                if len(cds) > 0:
                    cds = sorted(cds)
                    attr["cds_genome_start"] = cds[0].spanning_segment.start
                    attr["cds_genome_end"] = cds[-1].spanning_segment.end

                try:
                    my_tx = Transcript(*tuple(exon_segments + cds_segments), **attr)
                    if self.add_three_for_stop == True:
                        my_tx = add_three_for_stop_codon(my_tx)

                    transcripts.append(my_tx)

                except ValueError:
                    warn(
                        "Rejecting transcript '%s' because it contains exons "
                        " on multiple strands." % tname, DataWarning
                    )
                    # transcripts with exons on two strands
                    rejected.append(tname)
                except KeyError:
                    # transcripts where CDS ends outside bounds of transcript
                    # ideally this would not occur but does in rare cases
                    # e.g. there are 25 of these in flybase r5.43
                    warn(
                        "Rejecting transcript '%s because start or stop codons "
                        "are outside exon boundaries." % tname, DataWarning
                    )
                    rejected.append(tname)
            else:
                # transcript that has multiple subfeatures with shared ID but no children
                attr = self._tx_features[tname][0].attr
                attr["ID"] = tname
                attr["transcript_id"] = tname
                segments = [X.spanning_segment for X in self._tx_features[tname]]
                my_tx = Transcript(*tuple(segments), **attr)
                transcripts.append(my_tx)

            sys.exc_traceback = None

        return transcripts, rejected

    def _reset(self):
        """Release memory and reset internal hashes"""
        del self._feature_cache
        del self._tx_features
        gc.collect()
        del gc.garbage[:]
        self._tx_features = {}
        self._feature_cache = {X: copy.deepcopy({}) for X in self.transcript_components}
